"""Tauon Music Box - Basic Drawing and Text Drawing Functions Module"""

# Copyright © 2015-2019, Taiko2k captain(dot)gxj(at)gmail.com

#     This file is part of Tauon Music Box.
#
#     Tauon Music Box is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     Tauon Music Box is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Lesser General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with Tauon Music Box.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

import ctypes
import io
import logging
import math
from collections import OrderedDict
from ctypes import byref, c_bool, c_float, c_size_t
from typing import TYPE_CHECKING

import sdl3
from PIL import Image

from tauon.t_modules.t_extra import ColourRGBA, Timer, alpha_blend, coll_rect

if TYPE_CHECKING:
	from io import BytesIO

	from tauon.t_modules.t_main import Tauon

try:
	from jxlpy import JXLImagePlugin

	logging.info("Found jxlpy for JPEG XL support")
except ModuleNotFoundError:
	logging.warning("Unable to import jxlpy, JPEG XL support will be disabled.")
except Exception:
	logging.exception("Unknown error trying to import jxlpy, JPEG XL support will be disabled.")


import sys

if sys.platform == "win32":
	import os

	os.environ["PANGOCAIRO_BACKEND"] = "fc"

import cairo
import gi

gi.require_version("Pango", "1.0")
gi.require_version("PangoCairo", "1.0")
from gi.repository import Pango, PangoCairo


class QuickThumbnail:
	def __init__(self, tauon: Tauon) -> None:
		self.ddt = tauon.ddt
		self.renderer = tauon.renderer
		self.items: list[QuickThumbnail] = []
		self.queue: list[QuickThumbnail] = []
		self.rect = sdl3.SDL_FRect(0.0, 0.0)
		self.texture = None
		self.surface = None
		self.size = 50
		self.alive = False
		self.url = None

	def destruct(self) -> None:
		if self.surface:
			sdl3.SDL_DestroySurface(self.surface)
			self.surface = None
		if self.texture:
			sdl3.SDL_DestroyTexture(self.texture)
			self.texture = None
		self.alive = False

	def read_and_thumbnail(self, f: str, width: int, height: int) -> None:
		g = io.BytesIO()
		g.seek(0)
		im = Image.open(f)
		im.thumbnail((width, height), Image.Resampling.LANCZOS)
		im.save(g, "PNG")
		g.seek(0)
		self.surface = self.ddt.load_image(g)
		# self.items.append(self)
		self.alive = True

	def prime(self) -> None:
		texture = sdl3.SDL_CreateTextureFromSurface(self.renderer, self.surface)
		sdl3.SDL_DestroySurface(self.surface)
		self.surface = None
		tex_w = c_float(0.0)
		tex_h = c_float(0.0)
		sdl3.SDL_GetTextureSize(texture, byref(tex_w), byref(tex_h))
		self.rect.w = int(tex_w.value)
		self.rect.h = int(tex_h.value)
		self.texture = texture

	def draw(self, x: int, y: int) -> bool | None:
		if len(self.items) > 30:
			img = self.items[0]
			img.destruct()
			self.items.remove(img)
		if not self.alive:
			if self not in self.queue:
				self.queue.append(self)
			return False
		if not self.texture:
			self.prime()
		self.rect.x = round(x)
		self.rect.y = round(y)
		sdl3.SDL_RenderTexture(self.renderer, self.texture, None, self.rect)

		return True


perf = Timer()


class TDraw:
	def __init__(self, renderer: sdl3.LP_SDL_Renderer) -> None:
		# All
		self.renderer = renderer
		self.scale = 1
		self.force_subpixel_text = False

		# Drawing
		self.sdlrect = sdl3.SDL_FRect(10.0, 10.0, 10.0, 10.0)

		# Text and Fonts
		self.source_rect = sdl3.SDL_FRect(0.0, 0.0, 0.0, 0.0)
		self.dest_rect = sdl3.SDL_FRect(0.0, 0.0, 0.0, 0.0)

		self.surf = cairo.ImageSurface(cairo.FORMAT_ARGB32, 0, 0)
		self.context = cairo.Context(self.surf)
		self.layout = PangoCairo.create_layout(self.context)

		self.text_background_colour = ColourRGBA(0, 0, 0, 255)
		self.pretty_rect: tuple[int, int, int, int] | None = None
		self.real_bg: bool = False
		self.alpha_bg: bool = False
		self.force_gray: bool = False
		self.f_dict: dict[float, tuple[str, int, float]] = {}
		self.font_desc_cache: dict[float, Pango.FontDescription] = {}
		self.ttc: dict[
			tuple[int, str, int, int, int, int, int, int, int, int],
			list[sdl3.SDL_FRect | sdl3.LP_SDL_Texture | int | bool],
		] = {}
		self.ttl: list[tuple[int, str, int, int, int, int, int, int, int, int]] = []
		self.text_texture_cache_bytes = 0
		self.max_text_texture_cache_items = 96
		self.max_text_texture_cache_bytes = 12 * 1024 * 1024
		self.max_text_texture_cache_item_bytes = 256 * 1024
		self.text_wh_cache: OrderedDict[tuple[str, int, int, bool], tuple[int, int]] = OrderedDict()
		self.max_text_wh_cache_items = 2048

		self.was_truncated = False

	def load_image(self, g: BytesIO) -> sdl3.LP_SDL_Surface:
		size = g.getbuffer().nbytes
		pointer = ctypes.c_void_p(ctypes.addressof(ctypes.c_char.from_buffer(g.getbuffer())))
		stream = sdl3.SDL_IOFromMem(pointer, c_size_t(size))
		return sdl3.IMG_Load_IO(stream, c_bool(True))

	def rect_s(self, rectangle: tuple[int, int, int, int], colour: ColourRGBA, thickness: int) -> None:
		sdl3.SDL_SetRenderDrawColor(self.renderer, colour.r, colour.g, colour.b, colour.a)
		x, y, w, h = (round(x) for x in rectangle)
		th = math.floor(thickness)
		self.sdlrect.x = x - th
		self.sdlrect.y = y - th
		self.sdlrect.w = th
		self.sdlrect.h = h + th
		sdl3.SDL_RenderFillRect(self.renderer, self.sdlrect)  # left
		self.sdlrect.x = x - th
		self.sdlrect.y = y + h
		self.sdlrect.w = w + th
		self.sdlrect.h = th
		sdl3.SDL_RenderFillRect(self.renderer, self.sdlrect)  # bottom
		self.sdlrect.x = x
		self.sdlrect.y = y - th
		self.sdlrect.w = w + th
		self.sdlrect.h = th
		sdl3.SDL_RenderFillRect(self.renderer, self.sdlrect)  # top
		self.sdlrect.x = x + w
		self.sdlrect.y = y
		self.sdlrect.w = th
		self.sdlrect.h = h + th
		sdl3.SDL_RenderFillRect(self.renderer, self.sdlrect)  # right

	def rect_si(self, rectangle: tuple[int, int, int, int], colour: ColourRGBA, thickness: int) -> None:
		sdl3.SDL_SetRenderDrawColor(self.renderer, colour.r, colour.g, colour.b, colour.a)
		x, y, w, h = (round(x) for x in rectangle)
		th = math.floor(thickness)
		self.sdlrect.x = x
		self.sdlrect.y = y
		self.sdlrect.w = th
		self.sdlrect.h = h
		sdl3.SDL_RenderFillRect(self.renderer, self.sdlrect)  # left
		self.sdlrect.x = x
		self.sdlrect.y = y + (h - th)
		self.sdlrect.w = w
		self.sdlrect.h = th
		sdl3.SDL_RenderFillRect(self.renderer, self.sdlrect)  # bottom
		self.sdlrect.x = x
		self.sdlrect.y = y
		self.sdlrect.w = w
		self.sdlrect.h = th
		sdl3.SDL_RenderFillRect(self.renderer, self.sdlrect)  # top
		self.sdlrect.x = x + (w - th)
		self.sdlrect.y = y
		self.sdlrect.w = th
		self.sdlrect.h = h
		sdl3.SDL_RenderFillRect(self.renderer, self.sdlrect)  # right

	def rect_a(self, location_xy: list[int], size_wh: list[int], colour: ColourRGBA) -> None:
		self.rect((location_xy[0], location_xy[1], size_wh[0], size_wh[1]), colour)

	def clear_rect(self, rectangle: tuple[int, int, int, int]) -> None:
		sdl3.SDL_SetRenderDrawBlendMode(self.renderer, sdl3.SDL_BLENDMODE_NONE)
		sdl3.SDL_SetRenderDrawColor(self.renderer, 0, 0, 0, 0)

		self.sdlrect.x = float(rectangle[0])
		self.sdlrect.y = float(rectangle[1])
		self.sdlrect.w = float(rectangle[2])
		self.sdlrect.h = float(rectangle[3])

		sdl3.SDL_RenderFillRect(self.renderer, self.sdlrect)
		sdl3.SDL_SetRenderDrawBlendMode(self.renderer, sdl3.SDL_BLENDMODE_BLEND)

	def rect(self, rectangle: tuple[int, int, int, int], colour: ColourRGBA) -> None:
		sdl3.SDL_SetRenderDrawColor(self.renderer, colour.r, colour.g, colour.b, colour.a)

		self.sdlrect.x = float(rectangle[0])
		self.sdlrect.y = float(rectangle[1])
		self.sdlrect.w = float(rectangle[2])
		self.sdlrect.h = float(rectangle[3])

		# if fill:
		sdl3.SDL_RenderFillRect(self.renderer, self.sdlrect)
		# else:
		# 	sdl3.SDL_RenderDrawRect(self.renderer, self.sdlrect)

	def bordered_rect(
		self, rectangle: tuple[int, int, int, int], fill_colour: ColourRGBA, outer_colour: ColourRGBA, border_size: int
	) -> None:
		self.sdlrect.x = round(rectangle[0]) - border_size
		self.sdlrect.y = round(rectangle[1]) - border_size
		self.sdlrect.w = round(rectangle[2]) + border_size + border_size
		self.sdlrect.h = round(rectangle[3]) + border_size + border_size
		sdl3.SDL_SetRenderDrawColor(self.renderer, outer_colour.r, outer_colour.g, outer_colour.b, outer_colour.a)
		sdl3.SDL_RenderFillRect(self.renderer, self.sdlrect)
		self.sdlrect.x = round(rectangle[0])
		self.sdlrect.y = round(rectangle[1])
		self.sdlrect.w = round(rectangle[2])
		self.sdlrect.h = round(rectangle[3])
		sdl3.SDL_SetRenderDrawColor(self.renderer, fill_colour.r, fill_colour.g, fill_colour.b, fill_colour.a)
		sdl3.SDL_RenderFillRect(self.renderer, self.sdlrect)

	def line(self, x1: int, y1: int, x2: int, y2: int, colour: ColourRGBA) -> None:
		sdl3.SDL_SetRenderDrawColor(self.renderer, colour.r, colour.g, colour.b, colour.a)
		sdl3.SDL_RenderLine(self.renderer, round(x1), round(y1), round(x2), round(y2))

	def get_text_w(self, text: str, font: int, height: bool = False) -> int:
		x, y = self.get_text_wh(text, font, 3000)
		if height:
			return y
		return x

	def clear_text_cache(self) -> None:
		for key in self.ttl:
			so = self.ttc[key]
			sdl3.SDL_DestroyTexture(so[1])

		self.ttc.clear()
		self.ttl.clear()
		self.text_texture_cache_bytes = 0
		self.text_wh_cache.clear()

	def prime_font(self, name: str, size: float, user_handle: float, offset: int = 0) -> None:
		self.f_dict[user_handle] = (name + " " + str(size * self.scale), offset, size * self.scale)
		self.font_desc_cache[user_handle] = Pango.FontDescription(self.f_dict[user_handle][0])
		self.text_wh_cache.clear()

	def _font_description(self, font: int) -> Pango.FontDescription:
		cached = self.font_desc_cache.get(font)
		if cached is not None:
			return cached

		font_description = Pango.FontDescription(self.f_dict[font][0])
		self.font_desc_cache[font] = font_description
		return font_description

	def get_text_wh(self, text: str, font: int, max_x: int, wrap: bool = False) -> tuple[int, int] | None:
		cache_key = None
		if len(text) <= 256:
			cache_key = (text, font, max_x, wrap)
			cached = self.text_wh_cache.get(cache_key)
			if cached is not None:
				self.text_wh_cache.move_to_end(cache_key)
				return cached

		self.layout.set_font_description(self._font_description(font))
		self.layout.set_ellipsize(Pango.EllipsizeMode.END)
		self.layout.set_width(max_x * 1000)
		if wrap:
			self.layout.set_height(20000 * 1000)
		else:
			self.layout.set_height(0)

		try:
			self.layout.set_text(text, -1)
		except Exception:
			logging.exception(f"Exception in get_text_wh for: {text}")
			self.layout.set_text(text.encode("utf-8", "replace").decode("utf-8"), -1)

		result = self.layout.get_pixel_size()
		if cache_key is not None:
			self.text_wh_cache[cache_key] = result
			if len(self.text_wh_cache) > self.max_text_wh_cache_items:
				self.text_wh_cache.popitem(last=False)
		return result

	def get_y_offset(self, text: str, font: int, max_x: int, wrap: bool = False) -> int:
		"""HACKY"""
		self.layout.set_font_description(self._font_description(font))
		self.layout.set_ellipsize(Pango.EllipsizeMode.END)
		self.layout.set_width(max_x * 1000)
		if wrap:
			self.layout.set_height(20000 * 1000)
		else:
			self.layout.set_height(0)

		try:
			self.layout.set_text(text, -1)
		except Exception:
			logging.exception(f"Exception in get_y_offset for: {text}")
			self.layout.set_text(text.encode("utf-8", "replace").decode("utf-8"), -1)

		y_off = self.layout.get_baseline() / 1000
		y_off = round(round(y_off) - 13 * self.scale)  # 13 for compat with way text position used to work

		return y_off

	def __render_text(self, key: dict, x: int, y: int, range_top: int, range_height: int, align: int) -> None:
		sd = key

		if sd[3]:
			self.was_truncated = True

		if align == 1:
			sd[0].x = round(x) - sd[0].w

		elif align == 2:
			sd[0].x -= int(sd[0].w / 2)

		if range_height is not None and range_height < sd[0].h:
			if range_top < 0:
				range_top = 0

			if range_top > sd[0].h - range_height:
				range_top = sd[0].h - range_height

			self.source_rect.y = round(range_top)
			self.source_rect.w = sd[0].w
			self.source_rect.h = round(range_height)

			self.dest_rect.x = sd[0].x
			self.dest_rect.y = sd[0].y
			self.dest_rect.w = sd[0].w
			self.dest_rect.h = round(range_height)

			# sdl3.SDL_RenderCopyEx(self.renderer, sd[1], self.source_rect, self.dest_rect, 0, None, 0)
			sdl3.SDL_RenderTexture(self.renderer, sd[1], self.source_rect, self.dest_rect)
			return

		sdl3.SDL_RenderTexture(self.renderer, sd[1], None, sd[0])

	def __draw_text_cairo(
		self,
		location: list[int],
		text: str,
		colour: ColourRGBA,
		font: int,
		max_x: int,
		bg: ColourRGBA,
		align: int = 0,
		max_y: int | None = None,
		wrap: bool = False,
		range_top: int = 0,
		range_height: int | None = None,
		real_bg: bool = False,
		key: tuple[int, str, int, int, int, int, int, int, int, int] | None = None,
	) -> int:
		# perf.set()
		force_cache = False
		if key:
			force_cache = True

		self.was_truncated = False

		max_x += 12  # Hack
		max_x = round(max_x)

		alpha_bg = self.alpha_bg
		force_gray = self.force_gray
		# real_bg = True

		if bg.a < 200:
			alpha_bg = True
			force_gray = True

		x = round(location[0])
		y = round(location[1])

		if self.pretty_rect:
			w, h = self.get_text_wh(text, font, max_x, wrap)
			quick_box = [x, y, w, h]

			if align == 1:
				quick_box[0] = x - quick_box[2]
			elif align == 2:
				quick_box[0] -= int(quick_box[2] / 2)

			if coll_rect(self.pretty_rect, quick_box):
				# self.rect_r(quick_box, [0, 0, 0, 100], True)
				# if self.real_bg:
				# real_bg = True
				alpha_bg = True
			else:
				alpha_bg = False

		if alpha_bg:
			bg = ColourRGBA(0, 0, 0, 0)

		if max_y is not None:
			max_y = round(max_y)

		if len(text) == 0:
			return 0

		if key is None:
			key = (max_x, text, font, colour.r, colour.g, colour.b, colour.a, bg.r, bg.g, bg.b)

		if not real_bg or force_cache:
			sd = self.ttc.get(key)
			if sd:
				sd = self.ttc[key]
				sd[0].x = round(x)
				sd[0].y = round(y) - sd[2]

				self.__render_text(sd, x, y, range_top, range_height, align)
				self.ttl.remove(key)
				self.ttl.append(key)

				if wrap:
					return sd[0].h
				return sd[0].w

		if not self.pretty_rect:  # Would have already done this if True
			w, h = self.get_text_wh(text, font, max_x, wrap)

		if w < 1:
			return 0

		h += 4  # Compensate for characters that drop past the baseline, Pango doesn't seem to allow for this

		if wrap:
			w = max_x + 1

		if font not in self.f_dict:
			logging.info(f"Font not loaded: {font!s}")
			return 10

		format = sdl3.SDL_PIXELFORMAT_ARGB8888 if alpha_bg else sdl3.SDL_PIXELFORMAT_XRGB8888
		surface = None
		surface_locked = False
		pixel_data = None
		data = None
		surf = None
		context = None
		layout = None
		c = None
		y_off = 0

		try:
			if real_bg:
				box = sdl3.SDL_Rect(x, y - self.get_y_offset(text, font, max_x, wrap), w, h)

				if align == 1:
					box.x = x - box.w
				elif align == 2:
					box.x -= int(box.w / 2)

				ssurf = sdl3.SDL_RenderReadPixels(self.renderer, box)
				if not ssurf:
					logging.warning("SDL_RenderReadPixels failed while rendering text background")
					return 0

				if ssurf.contents.format != format:
					surface = sdl3.SDL_ConvertSurface(ssurf, format)
					sdl3.SDL_DestroySurface(ssurf)
					if not surface:
						logging.warning("SDL_ConvertSurface failed while rendering text background")
						return 0
				else:
					surface = ssurf
			else:
				surface = sdl3.SDL_CreateSurface(w, h, format)
				if not surface:
					logging.warning("SDL_CreateSurface failed while rendering text")
					return 0
				ctypes.memset(surface.contents.pixels, 0, surface.contents.pitch * h)

			if not sdl3.SDL_LockSurface(surface):
				logging.warning("SDL_LockSurface failed while rendering text")
				return 0
			surface_locked = True

			pixel_size = surface.contents.pitch * h
			pixel_data = (ctypes.c_ubyte * pixel_size).from_address(surface.contents.pixels)
			data = memoryview(pixel_data)

			if alpha_bg:
				surf = cairo.ImageSurface.create_for_data(data, cairo.FORMAT_ARGB32, w, h, surface.contents.pitch)
			else:
				surf = cairo.ImageSurface.create_for_data(data, cairo.FORMAT_RGB24, w, h, surface.contents.pitch)

			context = cairo.Context(surf)

			if force_gray:
				options = context.get_font_options()
				options.set_antialias(cairo.ANTIALIAS_GRAY)
				# options.set_hint_style(cairo.HINT_STYLE_NONE)
				context.set_font_options(options)
			elif self.force_subpixel_text:
				options = context.get_font_options()
				# options.set_antialias(cairo.ANTIALIAS_NONE)
				# options.set_antialias(cairo.ANTIALIAS_GRAY)
				options.set_antialias(cairo.ANTIALIAS_SUBPIXEL)
				context.set_font_options(options)

			layout = PangoCairo.create_layout(context)
			layout.set_auto_dir(False)

			if max_y is not None:
				layout.set_ellipsize(Pango.EllipsizeMode.END)
				layout.set_width(max_x * 1000)
				layout.set_height(max_y * 1000)
			else:
				layout.set_ellipsize(Pango.EllipsizeMode.END)
				layout.set_width(max_x * 1000)

				extra = 0
				if wrap:  # Compensate for height measurement being 1-2 lines too short. Pango bug?
					extra = round(400000 * self.scale)

				layout.set_height(h * 1000 + extra)

			if not wrap and max_y is None:
				layout.set_height(-1)

			# Attributes don't seem to be implemented in gi?
			# attrs = Pango.AttrList()
			# attrs.insert(Pango.Attribute(Pango.Underline.SINGLE))
			# layout.set_attributes(attrs)

			context.rectangle(0, 0, w, h)

			if not real_bg and not alpha_bg:
				context.set_source_rgb(bg.r / 255, bg.g / 255, bg.b / 255)
				# context.set_source_rgba(0, 0, 0, 0)
				context.fill()

			context.set_source_rgb(colour.r / 255, colour.g / 255, colour.b / 255)

			# desc = Pango.FontDescription(self.f_dict[font][0])
			# desc.set_family("Arial")

			layout.set_font_description(self._font_description(font))

			try:
				layout.set_text(text, -1)
			except Exception:
				logging.exception(f"Text error on text: {text}")
				layout.set_text(text.encode("utf-8", "replace").decode("utf-8"), -1)

			# logging.info(layout.get_direction(0))

			y_off = layout.get_baseline() / 1000
			y_off = round(round(y_off) - 13 * self.scale)  # 13 for compat with way text position used to work

			PangoCairo.show_layout(context, layout)

			self.was_truncated = layout.is_ellipsized()
			surf.flush()
			surf.finish()
			surf = None
			layout = None
			context = None
			data.release()
			data = None
			pixel_data = None
			sdl3.SDL_UnlockSurface(surface)
			surface_locked = False

			# Here the background colour is keyed out allowing lines to overlap slightly
			if not real_bg and not alpha_bg:
				format_details = sdl3.SDL_GetPixelFormatDetails(format)
				ke = sdl3.SDL_MapRGB(format_details, None, bg.r, bg.g, bg.b)
				sdl3.SDL_SetSurfaceColorKey(surface, True, ke)

			c = sdl3.SDL_CreateTextureFromSurface(self.renderer, surface)
			if not c:
				logging.warning("SDL_CreateTextureFromSurface failed while rendering text")
				return 0

			sdl3.SDL_DestroySurface(surface)
			surface = None

			if alpha_bg:
				blend_mode = sdl3.SDL_ComposeCustomBlendMode(
					sdl3.SDL_BLENDFACTOR_ONE,
					sdl3.SDL_BLENDFACTOR_ONE_MINUS_SRC_ALPHA,
					sdl3.SDL_BLENDOPERATION_ADD,
					sdl3.SDL_BLENDFACTOR_ONE,
					sdl3.SDL_BLENDFACTOR_ONE_MINUS_SRC_ALPHA,
					sdl3.SDL_BLENDOPERATION_ADD,
				)
				sdl3.SDL_SetTextureBlendMode(c, blend_mode)
		finally:
			layout = None
			context = None
			if surf is not None:
				surf.finish()
			if data is not None:
				data.release()
			pixel_data = None
			if surface_locked and surface is not None:
				sdl3.SDL_UnlockSurface(surface)
			if surface is not None:
				sdl3.SDL_DestroySurface(surface)

		dst = sdl3.SDL_FRect(round(x), round(y))
		dst.w = round(w)
		dst.h = round(h)
		dst.y = round(y) - y_off

		pack = [dst, c, y_off, self.was_truncated]
		cache_item_bytes = w * h * 4
		texture_cached = False

		try:
			self.__render_text(pack, x, y, range_top, range_height, align)

			# Don't cache if using real background data
			cache_texture = not real_bg and (
				force_cache or cache_item_bytes <= self.max_text_texture_cache_item_bytes
			)
			if cache_texture:
				pack.append(cache_item_bytes)
				self.ttc[key] = pack
				self.ttl.append(key)
				self.text_texture_cache_bytes += cache_item_bytes
				texture_cached = True
				while (
					len(self.ttl) > self.max_text_texture_cache_items
					or self.text_texture_cache_bytes > self.max_text_texture_cache_bytes
				):
					key = self.ttl[0]
					so = self.ttc[key]
					sdl3.SDL_DestroyTexture(so[1])
					if len(so) > 4:
						self.text_texture_cache_bytes -= so[4]
					del self.ttc[key]
					del self.ttl[0]
		finally:
			if c is not None and not texture_cached:
				sdl3.SDL_DestroyTexture(c)

		if wrap:
			return dst.h
		return dst.w

	def text(
		self,
		location: list[int],
		text: str,
		colour: ColourRGBA,
		font: int,
		max_w: int = 4000,
		bg: ColourRGBA | None = None,
		range_top: int = 0,
		range_height: int | None = None,
		real_bg: bool = False,
		key: tuple[int, str, int, int, int, int, int, int, int, int] | None = None,
	) -> int | None:
		# logging.info((text, font))

		if not text:
			return 0

		max_w = max(1, max_w)

		if bg is None:
			bg = self.text_background_colour

		if colour.a != 255:
			colour = alpha_blend(colour, bg)
		align = 0
		if len(location) > 2:
			if location[2] == 1:
				align = 1
			if location[2] == 2:
				align = 2
			if location[2] == 4:
				max_h = None
				if len(location) > 4:
					max_h = location[4]

				return self.__draw_text_cairo(
					location,
					text,
					colour,
					font,
					location[3],
					bg,
					max_y=max_h,
					wrap=True,
					range_top=range_top,
					range_height=range_height,
				)

		return self.__draw_text_cairo(location, text, colour, font, max_w, bg, align, real_bg=real_bg, key=key)



"""Tauon Music Box - Album chart image generator"""

# Copyright © 2015-2019, Taiko2k captain(dot)gxj(at)gmail.com

#     This file is part of Tauon Music Box.
#
#     Tauon Music Box is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     Tauon Music Box is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Lesser General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with Tauon Music Box.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING

import gi
from PIL import Image

gi.require_version("Pango", "1.0")
gi.require_version("PangoCairo", "1.0")
import cairo  # noqa: E402
from gi.repository import Pango, PangoCairo  # noqa: E402

if TYPE_CHECKING:
	from tauon.t_modules.t_main import Tauon, TrackClass


class TopChart:
	def __init__(self, tauon: Tauon) -> None:
		self.pctl = tauon.pctl
		self.cache_dir = tauon.cache_directory
		self.user_dir = tauon.user_directory
		self.album_art_gen = tauon.album_art_gen

	def generate(
		self,
		tracks: list[TrackClass],
		bg: tuple[int, int, int] = (10, 10, 10),
		rows: int = 3,
		columns: int = 3,
		show_lables: bool = True,
		font: str = "Monospace 10",
		tile: bool = False,
		cascade: tuple[tuple[int, int, int], tuple[int, int, int]] | None = None,
	) -> bool | str:
		# Main control variables
		border = 29
		text_offset = -7
		size = 170
		spacing = 9

		mode = 1
		if cascade:
			mode = 2

		if tile:
			border = 0
			spacing = 0
			size = 160
			text_offset = 15

		# Determine the final width and height of album grid
		h = round((border * 2) + (size * rows) + (spacing * (rows - 1)))
		w = round((border * 2) + (size * columns) + (spacing * (columns - 1)))

		if mode == 2:
			r1, r2, r3 = cascade[0]
			logging.info(r1 * 2 + r2 * 2 + r3 * 2)
			sets = []
			for q in range(100, 10000):
				a = q / r1
				b = q / r2
				c = q / r3

				if a - int(a) == b - int(b) == c - int(c) == 0:
					sets.append((int(a), int(b), int(c)))

			if not sets:
				return False

			abc = None
			for s in sets:
				if s[(r1, r2, r3).index(min((r1, r2, r3)))] > 165:
					abc = s
					break
			else:
				return False

			d1, d2, d3 = cascade[1]

			w = round(border * 2) + (abc[0] * r1)
			h = round(border * 2) + (abc[0] * d1) + (abc[1] * d2) + (abc[2] * d3)

		ww = w
		i = -1

		positions = []

		# Grid mode
		if mode == 1:
			for r in range(rows):
				for c in range(columns):
					i += 1

					# Break if we run out of albums
					if i > len(tracks) - 1:
						break

					# Determine coordinates for current album
					x = round(border + ((spacing + size) * c))
					y = round(border + ((spacing + size) * r))

					positions.append((tracks[i], x, y, size))

				positions.append(False)

		# Cascade mode
		elif mode == 2:
			a, b, c = abc

			size = a
			spacing = 0
			inv_space = 0
			if not tile:
				inv_space = 8

			x = border
			y = border

			for d in range(d1):
				for cl in range(r1):
					i += 1
					x = border + (spacing + size) * cl
					if i > len(tracks) - 1:
						break
					positions.append((tracks[i], x + inv_space // 2, y + inv_space // 2, size - inv_space))
				y += spacing + size

			size = b
			if not tile:
				inv_space = 6
			positions.append(False)

			for d in range(d2):
				for cl in range(r2):
					i += 1
					x = border + (spacing + size) * cl
					logging.info(x)
					if i > len(tracks) - 1:
						break
					positions.append((tracks[i], x + inv_space // 2, y + inv_space // 2, size - inv_space))
				y += spacing + size

			size = c
			if not tile:
				inv_space = 4
			positions.append(False)

			for d in range(d3):
				for cl in range(r3):
					i += 1
					x = border + (spacing + size) * cl
					if i > len(tracks) - 1:
						break
					positions.append((tracks[i], x + inv_space // 2, y + inv_space // 2, size - inv_space))
				y += spacing + size

		# Parse font
		font_comp = font.split(" ")
		font_size = int(font_comp.pop())

		col_w = 500

		two_col = False
		font_comp = font.split(" ")
		font_size = int(font_comp.pop())
		if len(positions) * (font_size + 4) > h - border * 2:
			two_col = True
			font_size += 1
			col_w = 380

		if show_lables:
			ww += col_w  # Add extra area to final size for text
			if two_col:
				ww += col_w + 20

		# Get labels
		text = ""
		b_text = ""
		f = False
		for item in positions:
			if item is False:
				if not f:
					text += " \n"  # Insert extra line to form groups for each row
				else:
					b_text += " \n"

				if two_col:
					f ^= True
				continue

			track, x, y, size = item

			# Determine the text label line
			artist = track.artist
			if track.album_artist:
				artist = track.album_artist
			line = f"{artist} - {track.album}\n"

			if not f:
				text += line
			else:
				b_text += line

		# Prepare a blank Cairo surface
		surface = cairo.ImageSurface(cairo.FORMAT_RGB24, ww, h)
		context = cairo.Context(surface)

		bg = (bg[0] / 255, bg[1] / 255, bg[2] / 255)  # Convert 8-bit rgb values to decimal
		context.set_source_rgb(bg[0], bg[1], bg[2])
		context.paint()  # Fill in the background colour

		for item in positions:
			if item is False:
				continue

			track, x, y, size = item

			# Export the album art to file object
			try:
				art_file = self.album_art_gen.save_thumb(track, (size, size), None, png=True, zoom=True)
			except Exception:
				logging.exception("Failed to save album art as file object")
				continue

			# Skip drawing this album if loading of artwork failed
			if not art_file:
				continue

			# Load image into Cairo and draw
			art = cairo.ImageSurface.create_from_png(art_file)
			context.set_source_surface(art, x, y)
			context.paint()

		if show_lables:
			# Setup font and prepare Pango layout
			options = context.get_font_options()
			options.set_antialias(cairo.ANTIALIAS_GRAY)
			context.set_font_options(options)
			layout = PangoCairo.create_layout(context)
			layout.set_ellipsize(Pango.EllipsizeMode.END)
			layout.set_width((col_w - text_offset - spacing) * 1000)
			# layout.set_height((h - spacing * 2) * 1000)
			# layout.set_spacing(3 * 1000)
			# layout.set_font_description(Pango.FontDescription(font))

			# Here we make sure the font size is small enough to fit
			font = " ".join(font_comp) + " " + str(font_size)
			layout.set_font_description(Pango.FontDescription(font))
			layout.set_text(text, -1)

			try:
				font_size = int(font_size)
				while font_size > 2:
					th = layout.get_pixel_size()[1]
					if th < h - border:
						break
					font_size -= 1
					font = " ".join(font_comp) + " " + str(font_size)
					layout.set_font_description(Pango.FontDescription(font))
					layout.set_text(text, -1)
			except Exception:
				logging.exception("Failed to adjust font size")

			# All good to render now
			y_text_padding = 3
			if tile:
				y_text_padding += 6
			context.translate(w + text_offset, border + y_text_padding)
			context.set_source_rgb(0.9, 0.9, 0.9)
			PangoCairo.show_layout(context, layout)

		if b_text:
			# Setup font and prepare Pango layout
			options = context.get_font_options()
			options.set_antialias(cairo.ANTIALIAS_GRAY)
			context.set_font_options(options)
			layout = PangoCairo.create_layout(context)
			layout.set_ellipsize(Pango.EllipsizeMode.END)
			layout.set_width((col_w - text_offset - spacing) * 1000)

			layout.set_font_description(Pango.FontDescription(font))
			layout.set_text(b_text, -1)

			# All good to render now
			y_text_padding = 3
			if tile:
				y_text_padding += 6
			context.translate(col_w + 10, 0)
			context.set_source_rgb(0.9, 0.9, 0.9)
			PangoCairo.show_layout(context, layout)

		# Finally export as PNG
		output_path = os.path.join(self.user_dir, "chart.png")
		surface.write_to_png(output_path)

		# Convert to JPEG for convenience
		output_path2 = os.path.join(self.user_dir, "chart.jpg")
		im = Image.open(output_path)
		im.save(output_path2, "JPEG", quality=92)

		return output_path2



"""Tauon Music Box - URL stream download and encoding module"""

# Copyright © 2020, Taiko2k captain(dot)gxj(at)gmail.com

#     This file is part of Tauon Music Box.
#
#     Tauon Music Box is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     Tauon Music Box is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Lesser General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with Tauon Music Box.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

import copy
import datetime
import io
import logging
import os
import shutil
import subprocess
import sys
import threading
import time
import urllib.request

import mutagen

from tauon.t_modules.t_enums import Backend
from tauon.t_modules.t_extra import filename_safe

if sys.platform != "win32":
	import fcntl
from typing import TYPE_CHECKING

if TYPE_CHECKING:
	from subprocess import Popen
	from urllib.request import _UrlopenRet

	from tauon.t_modules.t_main import Tauon


class StreamEnc:
	def __init__(self, tauon: Tauon) -> None:
		self.tauon = tauon
		self.download_running = False
		self.encode_running = False
		self.pump_running = False
		self.feed_running = False

		self.download_process = False
		self.abort = False

		self.s_name = ""
		self.s_bitrate = ""
		self.s_genre = ""
		self.s_description = ""
		self.s_mime = ""
		self.s_format = ""

		self.chunks = {}
		self.c = 0
		self.url = ""

	def stop(self) -> None:
		if self.tauon.radiobox.websocket:
			self.tauon.radiobox.websocket.close()
			logging.info("Websocket closed")

		self.abort = True
		self.tauon.radiobox.loaded_url = None

	def start_download(self, url: str) -> bool:
		self.abort = True
		while self.download_running:
			time.sleep(0.01)
		while self.encode_running:
			time.sleep(0.01)
		while self.pump_running:
			time.sleep(0.01)

		self.__init__(self.tauon)

		self.url = url
		result = self.start_request()
		if not result:
			return False

		self.download_process = threading.Thread(target=self.pump)
		self.download_process.daemon = True
		self.download_process.start()
		return True

	def start_request(self) -> bool:
		url = self.url

		def NiceToICY(self) -> None:
			class InterceptedHTTPResponse:
				pass

			if not url.endswith(".ts"):
				line = self.fp.readline().replace(b"ICY 200 OK\r\n", b"HTTP/1.0 200 OK\r\n")
			else:
				line = self.fp.readline()
			InterceptedSelf = InterceptedHTTPResponse()
			InterceptedSelf.fp = io.BufferedReader(io.BytesIO(line))
			InterceptedSelf.debuglevel = self.debuglevel
			InterceptedSelf._close_conn = self._close_conn
			return ORIGINAL_HTTP_CLIENT_READ_STATUS(InterceptedSelf)

		ORIGINAL_HTTP_CLIENT_READ_STATUS = urllib.request.http.client.HTTPResponse._read_status
		urllib.request.http.client.HTTPResponse._read_status = NiceToICY

		retry = 5
		while True:
			try:
				urllib.request.http.client.HTTPResponse._read_status = NiceToICY
				r = urllib.request.Request(self.url)
				# r.add_header('GET', '1')
				if not self.url.endswith(".ts"):
					r.add_header("Icy-MetaData", "1")
				r.add_header("User-Agent", self.tauon.t_agent)
				logging.info("Open URL.....")
				r = urllib.request.urlopen(r, timeout=20, context=self.tauon.tls_context)
				logging.info("URL opened.")

			except Exception as e:
				# TODO(Martin): Specify the exception better and turn the top part into debug statements, then only throw except when Connection fails below
				logging.exception("URL error...")
				retry -= 1
				if retry > 0 and "Temporary" in str(e):
					time.sleep(2)
					logging.debug("RETRYING...")
					continue
				logging.error("Connection failed")
				self.tauon.show_message(_("Failed to establish a connection"), str(e), mode="error")
				return False
			finally:
				urllib.request.http.client.HTTPResponse._read_status = ORIGINAL_HTTP_CLIENT_READ_STATUS
			break

		self.download_process = threading.Thread(target=self.run_download, args=([r]))
		self.download_process.daemon = True
		self.download_process.start()
		self.download_running = True
		return True

	def pump(self) -> None:
		aud = self.tauon.aud
		if self.tauon.prefs.backend != Backend.PHAZOR or not aud:
			logging.error("Radio error: Phazor not loaded")
			return
		self.pump_running = True

		rate = str(self.tauon.prefs.samplerate)
		# fmt:off
		cmd = [
			str(self.tauon.get_ffmpeg()),
			"-loglevel", "quiet",
			"-i", "pipe:0",
			"-acodec", "pcm_s16le",
			"-f", "s16le",
			"-ac", "2",
			"-ar", rate,
			"-",
		]
		# fmt:on

		startupinfo = None
		if self.tauon.windows:
			startupinfo = subprocess.STARTUPINFO()
			startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
		decoder = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, startupinfo=startupinfo)
		if sys.platform != "win32":
			fcntl.fcntl(decoder.stdout.fileno(), fcntl.F_SETFL, os.O_NONBLOCK)

		raw_audio = None
		max_read = 10000
		self.tauon.vb.reset()

		def feed(decoder: Popen[bytes]) -> None:
			position = 0
			self.feed_running = True
			try:
				while True:
					if position < self.tauon.stream_proxy.c:
						if position not in self.tauon.stream_proxy.chunks:
							logging.info("The buffer was deleted too soon!")
							break

						chunk = self.chunks[position]
						decoder.stdin.write(chunk)
						self.tauon.vb.input(self.tauon.stream_proxy.chunks[position])
						position += 1
					else:
						time.sleep(0.01)
					if self.abort or not self.pump_running or not self.feed_running:
						break
			except Exception:
				logging.exception("Feed not running!")
				self.feed_running = False
				raise
			logging.info("Exit feeder")

		feeder = threading.Thread(target=feed, args=[decoder])
		feeder.daemon = True
		feeder.start()

		retry = 3

		while True:
			if not self.tauon.stream_proxy.download_running or self.abort:
				break
			if raw_audio is None:
				raw_audio = decoder.stdout.read(max_read)
			if raw_audio:
				r = aud.feed_ready(max_read)
				if r:
					aud.feed_raw(len(raw_audio), raw_audio)
					if len(raw_audio) < max_read:
						time.sleep(0.01)
					raw_audio = None
					continue

			time.sleep(0.01)

		decoder.terminate()
		time.sleep(0.1)
		try:
			decoder.kill()
		except Exception:
			logging.exception("Failed to kill decoder")

		self.pump_running = False

	def encode(self) -> None:
		self.encode_running = True

		try:
			ffmpeg_path = self.tauon.get_ffmpeg()
			if ffmpeg_path is None:
				logging.error("FFmpeg could not be found for stream encoder")
				self.encode_running = False
				return

			while self.c < 20:
				if self.abort:
					self.encode_running = False
					return
				time.sleep(0.05)

			ext = ".opus"
			rate = "48000"
			codec = self.tauon.prefs.radio_record_codec.upper()
			if codec == "OGG":
				ext = ".ogg"
				rate = "44100"
			if codec == "MP3":
				ext = ".mp3"
				rate = "44100"
			if codec == "FLAC":
				ext = ".flac"
				rate = "44100"

			target_file = self.tauon.cache_directory / "stream" / ext
			target_file.parent.mkdir(parents=True, exist_ok=True)
			if target_file.is_file():
				target_file.unlink()

			# fmt:off
			cmd = [
				str(ffmpeg_path),
				"-loglevel", "quiet",
				"-i", "pipe:0",
				"-acodec", "pcm_s16le",
				"-f", "s16le",
				"-ac", "2",
				"-ar", rate,
				"-",
			]
			# fmt:on

			decoder = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
			if sys.platform != "win32":
				fcntl.fcntl(decoder.stdout.fileno(), fcntl.F_SETFL, os.O_NONBLOCK)

			position = 0
			old_metadata = self.tauon.radiobox.song_key
			old_tags = self.tauon.pctl.found_tags

			##cmd = ["opusenc", "--raw", "--raw-rate", "48000", "-", target_file]
			# fmt:off
			cmd = [
				str(ffmpeg_path),
				"-loglevel", "quiet",
				"-f", "s16le",
				"-ar", rate,
				"-ac", "2",
				"-i", "pipe:0",
				str(target_file),
			]
			# fmt:on
			encoder = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE)

			def save_track() -> None:
				# self.tauon.recorded_songs.append(song)

				save_file = f"{datetime.datetime.now():%Y-%m-%d %H-%M-%S} - "
				save_file += filename_safe(old_metadata)
				save_file = save_file.strip() + ext
				save_file = os.path.join(self.tauon.prefs.encoder_output, save_file)
				if os.path.exists(save_file):
					os.remove(save_file)
				if not os.path.exists(self.tauon.prefs.encoder_output):
					os.makedirs(self.tauon.prefs.encoder_output)
				shutil.move(target_file, save_file)

				# logging.info(self.tauon.pctl.tag_history)
				# logging.info(old_metadata)
				tags = self.tauon.pctl.tag_history.get(old_metadata, None)
				if tags:
					logging.info("Save metadata to file")
					# logging.info(tags)
					muta = mutagen.File(save_file, easy=True)
					muta["artist"] = tags.get("artist", "")
					muta["title"] = tags.get("title", "")
					muta["album"] = tags.get("album", "")
					# if tags["image"]:
					# 	tags["image"].seek(0)
					# 	im = Image.open(tags["image"])
					# 	width, height = im.size
					# 	tags["image"].seek(0)
					#
					# 	picture = Picture()
					# 	tags["image"].seek(0)
					# 	picture.data = tags["image"].read()
					# 	picture.type = 3
					# 	picture.desc = ""
					# 	picture.mime = "image/jpeg"
					# 	picture.width = width
					# 	picture.height = height
					#
					# 	mode_to_bpp = {'1': 1, 'L': 8, 'P': 8, 'RGB': 24, 'RGBA': 32, 'CMYK': 32, 'YCbCr': 24, 'I': 32,
					# 	'F': 32}
					# 	picture.depth = mode_to_bpp[im.mode]
					#
					# 	picture_data = picture.write()
					# 	encoded_data = base64.b64encode(picture_data)
					# 	vcomment_value = encoded_data.decode("ascii")
					# 	muta["metadata_block_picture"] = [vcomment_value]

					muta.save()

				target_pl = None
				for i, pl in enumerate(self.tauon.pctl.multi_playlist):
					if pl.title == "Saved Radio Tracks":
						target_pl = i
				if target_pl is None:
					self.tauon.pctl.multi_playlist.append(self.tauon.pl_gen(title="Saved Radio Tracks"))
					target_pl = len(self.tauon.pctl.multi_playlist) - 1

				load_order = self.tauon.pctl.LoadClass()
				load_order.playlist = self.tauon.pctl.multi_playlist[target_pl].uuid_int
				load_order.target = save_file
				self.tauon.load_orders.append(copy.deepcopy(load_order))
				self.tauon.gui.update += 1

			while True:
				if self.abort:
					decoder.terminate()
					encoder.terminate()
					time.sleep(0.1)
					try:
						decoder.kill()
					except Exception:
						logging.exception("Failed to kill decoder")
					try:
						encoder.kill()
					except Exception:
						logging.exception("Failed to kill encoder")

					if target_file.exists():
						if target_file.stat().st_size > 256000:
							logging.info("Save file")
							save_track()
						else:
							logging.info("Discard small file")
							target_file.unlink()

					self.encode_running = False
					self.tauon.pctl.tag_history.clear()
					return

				if old_metadata != self.tauon.radiobox.song_key:
					if (
						(self.c < 400 and not old_metadata)
						or not target_file.exists()
						or target_file.stat().st_size < 100000
					):
						old_metadata = self.tauon.radiobox.song_key
					else:
						logging.info("Split and save file")
						encoder.stdin.close()
						try:
							encoder.wait(timeout=4)
						except Exception:
							logging.exception("Encoder timed out")
						try:
							encoder.kill()
						except Exception:
							logging.exception("Failed to kill encoder")
						if target_file.exists():
							if target_file.stat().st_size > 256000:
								save_track()
							else:
								logging.info("Discard small file")
								target_file.unlink()
						encoder = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE)

				raw_audio = decoder.stdout.read(1000000)
				if raw_audio:
					encoder.stdin.write(raw_audio)

				if position < self.c:
					chunk = self.chunks[position]
					position += 1
					decoder.stdin.write(chunk)
				else:
					time.sleep(0.005)

		except Exception:
			logging.exception("Encoder thread crashed!")
			self.encode_running = False
			return

	def run_download(self, r: _UrlopenRet) -> None:
		h = r.info()

		self.s_name = h.get("icy-name")
		metaint = h.get("icy-metaint")
		self.s_bitrate = h.get("icy-br")
		self.s_genre = h.get("icy-genre")
		self.s_description = h.get("icy-description")
		self.s_mime = h.get("Content-Type")

		logging.info(self.s_mime)
		if self.s_mime == "audio/mpeg":
			self.s_format = "MP3"
		if self.s_mime == "audio/ogg":
			self.s_format = "OGG"
		if self.s_mime == "audio/aac":
			self.s_format = "AAC"
		if self.s_mime == "audio/aacp":
			self.s_format = "AAC+"

		test_done = 0

		icy = False
		m_remain = 0
		m = 0
		if metaint and int(metaint) > 0:
			m = int(metaint)
			m_remain = m
			icy = True

		maybe = b""

		if self.tauon.prefs.auto_rec:
			self.download_process = threading.Thread(target=self.encode)
			self.download_process.daemon = True
			self.download_process.start()

		try:
			while True:
				chunk = r.read(256)

				if self.abort:
					r.close()
					logging.info("Abort stream connection")
					self.download_running = False
					return
				if not chunk:
					r.close()
					logging.info("Stream connection closed")
					self.download_running = False
					self.abort = True
					return

				if chunk:
					if not icy or m_remain > len(chunk):
						# We're sure its data Its data, send it on
						self.chunks[self.c] = chunk

						# Delete old data
						d = self.c - 90000
						if d in self.chunks:
							del self.chunks[d]

						test_done += len(chunk)
						self.c += 1
						m_remain -= len(chunk)

						continue
					# It may contain the metadata block, put it aside
					maybe += chunk

				# Try to extract ICY tag
				if maybe:
					data1 = maybe[:m_remain]
					inter = maybe[m_remain:]

					# Read the metadata length byte
					if inter:
						special = inter[0]
						follow = special * 16

						if len(inter) < follow + 2:
							# Not enough data yet
							continue

						text = inter[1 : follow + 1]
						data2 = inter[follow + 1 :]

						self.chunks[self.c] = data1 + data2
						# Delete old data
						d = self.c - 90000
						if d in self.chunks:
							del self.chunks[d]

						self.c += 1

						test_done += len(data1)
						test_done = 0

						m_remain = m - len(data2)
						test_done += len(data2)
						maybe = b""

						try:
							meta = text.decode().rstrip("\x00")
							for tag in meta.split(";"):
								if "=" in tag:
									a, b = tag.split("=", 1)
									if a == "StreamTitle":
										# logging.info("Set meta")
										self.tauon.pctl.tag_meta = b.rstrip("'").lstrip("'")
										break
						except Exception:
							logging.exception("Data malformation detected. Stream aborted.")
							r.close()
							self.download_running = False
							self.abort = True
							self.tauon.show_message(_("Data malformation detected. Stream aborted."), mode="error")
							raise
		except Exception:
			logging.exception("Stream download thread crashed!")
			self.download_running = False
			self.abort = True
			return



"""Tauon Music Box - Phazor audio backend module"""

# Copyright © 2020, Taiko2k captain(dot)gxj(at)gmail.com

#     This file is part of Tauon Music Box.
#
#     Tauon Music Box is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     Tauon Music Box is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Lesser General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with Tauon Music Box.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import ctypes
import logging
import sys
from http import HTTPStatus

if sys.platform == "win32":
	from ctypes import WINFUNCTYPE
import hashlib
import importlib.machinery
import math
import os.path
import shutil
import subprocess
import sysconfig
import threading
import time
from ctypes import CFUNCTYPE, POINTER, c_char, c_char_p, c_int, c_void_p, cast
from pathlib import Path
from typing import TYPE_CHECKING

import requests
from requests.models import PreparedRequest

from tauon.t_modules.t_enums import PlayerState, PlayingState
from tauon.t_modules.t_extra import Timer, shooter, tmp_cache_dir

if TYPE_CHECKING:
	from ctypes import CDLL
	from subprocess import Popen

	from tauon.t_modules.t_main import GuiVar, PlayerCtl, Tauon, TrackClass
	from tauon.t_modules.t_prefs import Prefs
	from tauon.t_modules.t_spot import SpotCtl


class LibreSpot:
	def __init__(self, tauon: Tauon) -> None:
		# fmt:off
		self.tauon:      Tauon = tauon
		self.aud:         CDLL = tauon.aud
		self.gui:       GuiVar = tauon.gui
		self.windows:        bool = tauon.windows
		self.pctl:   PlayerCtl = tauon.pctl
		self.prefs:      Prefs = tauon.prefs
		self.spot_ctl: SpotCtl = tauon.spot_ctl
		self.librespot_p: Popen[bytes] | None = tauon.librespot_p
		self.show_message    = tauon.show_message
		self.cache_directory: Path = tauon.cache_directory
		self.running:   bool = False
		self.flush:     bool = False
		# fmt:on

	def go(self, force: bool = False) -> int:
		self.aud.config_set_feed_samplerate(44100)
		self.aud.config_set_min_buffer(1000)
		if not shutil.which("librespot"):
			self.show_message(_("SPP: Error, librespot not found"))
			return 1
		# if not prefs.spot_username or not prefs.spot_password:
		# 	self.show_message("Please enter your spotify username and password in settings")
		# 	return 1
		if self.librespot_p:
			if force:
				logging.info("SPP: Force restart librespot")
				self.end()
				self.librespot_p = None
			else:
				self.flush = True

		self.pctl.spot_playing = True

		if not self.librespot_p:
			logging.info("SPP: Librespot not running")
			username = self.spot_ctl.get_username()
			if not username or not self.prefs.spotify_token:
				logging.error("SPP: Missing auth data")
				return 1

			cache = str(self.cache_directory / "lsspot")
			if not Path(cache).exists():
				os.makedirs(cache)

			access_token = str(self.spot_ctl.spotify.token.access_token)

			cmd = [
				"librespot",
				"-k",
				access_token,
				"--backend",
				"pipe",
				"-n",
				"Tauon",
				"--disable-audio-cache",
				"--device-type",
				"computer",
				"--volume-ctrl",
				"fixed",
				"--initial-volume",
				"100",
			]

			# self.spot_ctl.preparing_spotify = True
			self.gui.update += 1

			startupinfo = None
			if sys.platform == "win32":
				startupinfo = subprocess.STARTUPINFO()
				startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
			try:
				self.librespot_p = subprocess.Popen(cmd, stdout=subprocess.PIPE, startupinfo=startupinfo)
				logging.info("SPP: Librespot now started")
			except Exception:
				logging.exception("SPP: Error starting librespot")
				# tauon.spot_ctl.preparing_spotify = False
				self.gui.update += 1
				return 1
		return 0

	def soft_end(self) -> None:
		self.running = False
		self.pctl.spot_playing = False

	def end(self) -> None:
		self.running = False
		self.pctl.spot_playing = False
		if self.librespot_p:
			if self.windows:
				self.librespot_p.terminate()
				# self.librespot_p.communicate()
			else:
				import signal

				self.librespot_p.send_signal(signal.SIGINT)  # terminate() doesn't work

	def worker(self) -> None:
		self.running = True
		logging.info("SPP: Enter librespot feeder thread")
		while True:
			if self.running is False:
				logging.info("SPP: Exit Librespot worker thread")
				return
			if self.flush:
				self.flush = False
				logging.info("SPP: Flushing some data...")
				if self.librespot_p is not None:
					self.librespot_p.stdout.read(70000)  # rough
				logging.info("SPP: Done flush")
			if self.aud.feed_ready(1000) == 1 and self.aud.get_buff_fill() < 2000 and self.librespot_p is not None:
				data = self.librespot_p.stdout.read(1000)
				self.aud.feed_raw(len(data), data)

			if self.tauon.player4_state == PlayerState.STOPPED:
				if self.librespot_p is not None:
					self.librespot_p.stdout.read(50)
				time.sleep(0.0002)
			else:
				time.sleep(0.002)


class FFRun:
	def __init__(self, tauon: Tauon) -> None:
		self.tauon: Tauon = tauon
		self.decoder = None

	def close(self) -> None:
		if self.decoder:
			self.decoder.terminate()
			if self.decoder.stdin:
				logging.debug("Closing STDIN in FFrun")
				self.decoder.stdin.close()
			if self.decoder.stdout:
				logging.debug("Closing STDOUT in FFrun")
				self.decoder.stdout.close()
			if self.decoder.stderr:
				logging.debug("Closing STDERR in FFrun")
				self.decoder.stderr.close()
			self.decoder.wait()  # Ensure the process fully terminates
		self.decoder = None

	def start(self, uri: bytes, start_ms: int, samplerate: int) -> int:
		self.close()
		ffmpeg_path = self.tauon.get_ffmpeg()
		if ffmpeg_path is None:
			self.tauon.test_ffmpeg()
			return 1
		path = str(ffmpeg_path)
		cmd = [path]
		cmd += ["-loglevel", "quiet"]
		if start_ms > 0:
			cmd += ["-ss", f"{start_ms}ms"]
		cmd += ["-i", uri.decode(), "-acodec", "pcm_s16le", "-f", "s16le", "-ac", "2", "-ar", f"{samplerate}", "-"]
		startupinfo = None
		if sys.platform == "win32":
			startupinfo = subprocess.STARTUPINFO()
			startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
		try:
			self.decoder = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, startupinfo=startupinfo)
		except Exception:
			logging.exception("Failed to start FFmpeg")
			return 1
		return 0

	def read(self, buffer: int, maximum: int) -> int:
		if self.decoder:
			data = self.decoder.stdout.read(maximum)
			p = cast(buffer, POINTER(c_char * maximum))
			p.contents.value = data
			return len(data)
		return 0


class Cachement:
	def __init__(self, tauon: Tauon) -> None:
		# fmt:off
		self.tauon        = tauon
		self.gui          = tauon.gui
		self.pctl         = tauon.pctl
		self.prefs        = tauon.prefs
		self.show_message = tauon.show_message
		self.audio_cache  = tauon.cache_directory / "network-audio1"
		self.audio_cache2 = tauon.cache_directory / "audio-cache"
		# fmt:on
		self.direc = str(self.audio_cache2)
		if self.prefs.tmp_cache:
			self.direc = os.path.join(tmp_cache_dir(), "audio-cache")
		if not Path(self.direc).exists():
			os.makedirs(self.direc)
		self.list: list[str] = self.prefs.cache_list
		self.files = os.listdir(self.direc)
		self.get_now = None
		self.running = False
		self.ready = None
		self.error = None

	def get_key(self, track: TrackClass) -> str:
		if track.is_network:
			return hashlib.sha256((str(track.index) + track.url_key).encode()).hexdigest()
		return hashlib.sha256(track.fullpath.encode()).hexdigest()

	def get_file_cached_only(self, track: TrackClass) -> str | None:
		key = self.get_key(track)
		if key in self.files:
			path = Path(self.direc) / key
			if path.is_file():
				return str(path)
		return None

	def get_file(self, track: TrackClass) -> tuple[int, str | None]:
		# 0: file ready
		# 1: file downloading
		# 2: file not found
		if self.error == track:
			return 2, None

		key = self.get_key(track)
		path = os.path.join(self.direc, key)

		if self.running and self.get_now == track:
			return 1, path

		if key in self.files and os.path.isfile(path):
			logging.info("Got cached file")
			self.files.remove(key)
			self.files.append(key)  # bump to top of list
			self.get_now = None
			if not self.running:
				shoot_dl = threading.Thread(target=self.run)
				shoot_dl.daemon = True
				shoot_dl.start()
			return 0, path

		# disable me for debugging
		for codec in (".opus", ".ogg", ".flac", ".mp3"):
			idea = (
				os.path.join(
					self.prefs.encoder_output, self.tauon.encode_folder_name(track), self.tauon.encode_track_name(track)
				)
				+ codec
			)
			if os.path.isfile(idea):
				logging.info("Found transcode")
				return 0, idea

		self.get_now = track
		if not self.running:
			shoot_dl = threading.Thread(target=self.run)
			shoot_dl.daemon = True
			shoot_dl.start()
		return 1, path

	def run(self) -> None:
		self.running = True

		now = self.get_now
		self.get_now = None
		if now is not None:
			error = self.dl_file(now)
			if error:
				self.error = now
				self.running = False
				return

		if self.get_now is None:
			i = 0
			while i < 10:
				time.sleep(0.1)
				i += 1
				if self.get_now is not None:
					self.running = False
					return
			logging.info("Precache next track")
			next_track = self.pctl.advance(dry=True)
			if next_track is not None:
				self.dl_file(self.pctl.get_track(next_track))

		self.trim_cache()
		self.running = False
		return

	def trim_cache(self) -> None:
		# Remove untracked items
		for item in self.files:
			t = os.path.join(self.direc, item)
			if os.path.isfile(t) and item not in self.list:
				os.remove(t)

		# Check total size
		limit = self.prefs.cache_limit
		if self.prefs.tmp_cache:
			limit = 10
		while True:
			s = 0
			for item in list(self.list):
				t = os.path.join(self.direc, item)
				if not os.path.exists(t):
					self.list.remove(item)
					continue
				s += os.path.getsize(t)
			# Removed oldest items if over limit
			if s > limit * 1000 * 1000 and len(self.list) > 3:
				t = os.path.join(self.direc, self.list[0])
				os.remove(t)
				del self.list[0]
			else:
				break

	def dl_file(self, track: TrackClass) -> int | None:
		self.pctl.buffering_percent = 0
		key = self.get_key(track)
		path = Path(self.direc) / key
		if path.exists():
			if not path.is_file():
				return 1
			if key in self.list:
				return 0
			path.unlink()
		if not track.is_network:
			if not Path(track.fullpath).is_file():
				self.error = track
				self.running = False
				return 1

			logging.info("Start transfer")
			timer = Timer()
			target = path.open("wb")
			source = Path(track.fullpath).open("rb")
			while True:
				try:
					data = source.read(128000)
				except Exception:
					logging.exception("Transfer failed.")
					break
				if len(data) > 0:
					logging.info(f"Caching file @ {int(len(data) / timer.hit() / 1000)} kbps")
				else:
					break
				target.write(data)
			target.close()
			source.close()
			logging.info("got file")
			self.files.append(key)
			self.list.append(key)
			return 0

		try:
			logging.info("Download file")

			network_url, params = self.pctl.get_url(track)
			if not network_url:
				logging.info("No URL")
				return 1
			if type(network_url) in (list, tuple) and len(network_url) == 1:
				network_url = network_url[0]
			elif type(network_url) in (list, tuple):
				logging.info("Multi part DL")
				logging.info(path)
				# fmt:off
				ffmpeg_command = [
					str(self.tauon.get_ffmpeg()),
					"-i", "-",      # Input from stdin (pipe the data)
					"-f", "flac",   # Specify FLAC as the output format explicitly
					"-c:a", "copy", # Copy FLAC data without re-encoding
					path,           # Output file for extracted FLAC
				]
				# fmt:on
				p = subprocess.Popen(ffmpeg_command, stdin=subprocess.PIPE)
				i = 0
				for url in network_url:
					i += 1
					logging.info(i)
					response = requests.get(url, timeout=10)
					if response.status_code == HTTPStatus.OK:
						p.stdin.write(response.content)
					else:
						logging.error(f"ERROR CODE: {response.status_code}")
					if i == 3:
						self.ready = track

					self.gui.update += 1
					self.pctl.buffering_percent = math.floor(i / len(network_url) * 100)
					self.gui.buffering_text = str(self.pctl.buffering_percent) + "%"
					if self.get_now is not None and self.get_now != track:
						logging.info("Aborted loading track!")
						return None

				# logging.info("done")
				p.stdin.close()
				p.wait()

				logging.info("Done loading track")
				# self.ready = track
				self.files.append(key)
				self.list.append(key)
				return None

			part = requests.get(network_url, stream=True, params=params, timeout=(3, 10))

			if part.status_code == HTTPStatus.NOT_FOUND:
				self.show_message("Server: File not found", mode="error")
				self.error = track
				return 1
			if part.status_code != HTTPStatus.OK:
				self.show_message("Server Error", mode="error")
				self.error = track
				return 1

		except Exception as e:
			logging.exception("Download failed!")
			self.show_message(_("Error"), str(e), mode="error")
			self.error = track
			return 1

		a = 0
		length = 0
		cl = part.headers.get("Content-Length")
		if cl:
			length = int(cl)
			self.gui.buffering_text = "0%"

		timer = Timer()
		try:
			with path.open("wb") as f:
				for chunk in part.iter_content(chunk_size=1024):
					if chunk:  # filter out keep-alive new chunks
						a += 1
						if a == 3000 and self.prefs.jump_start:  # kilobyes~
							self.ready = track
						if a % 32 == 0:
							# time.sleep(0.03)
							# logging.info(f"Downloading file @ {round(32 / timer.hit())} kbps")
							if length:
								self.gui.update += 1
								self.pctl.buffering_percent = round(a * 1000 / length * 100)
								self.gui.buffering_text = str(round(a * 1000 / length * 100)) + "%"

						if self.get_now is not None and self.get_now != track:
							logging.warning("Aborted loading track!")
							return None

						# if self.cancel is True:
						# 	self.part.close()
						# 	self.status = "failed"
						# 	logging.info("Abort download")
						# 	return

						f.write(chunk)
			logging.info("Done loading track")
			self.ready = track
			self.files.append(key)
			self.list.append(key)
		except Exception:
			logging.exception("Download failed!")
			return 1
		return 0


def find_library(libname: str) -> Path | None:
	"""Search for 'libname.extension' in various formats.

	Return library Path loadable with ctypes.CDLL, otherwise return None
	"""
	# Can look like ~/Projects/Tauon/t_modules
	base_path = Path(__file__).parent
	search_paths: list[Path] = []
	so_extensions = importlib.machinery.EXTENSION_SUFFIXES
	site_packages_path = sysconfig.get_path("purelib")

	# Used to be lib but Slackware uses that for 32-bit,
	# all other distros seem to symlink lib64 to lib, so just use lib64
	libdir = "lib64"

	# Try looking in site-packages of the current environment, pwd and ../pwd
	for extension in so_extensions:
		# fmt:off
		search_paths += [
			(Path(site_packages_path)                        / (libname + extension)).resolve(),
			(Path(base_path)          / ".."                 / (libname + extension)).resolve(),
			(Path(base_path)          / ".." / ".."          / (libname + extension)).resolve(),
			# Compat with old way to store .so files
			(Path(base_path)          / ".." / ".." / libdir / (libname + ".so")).resolve(),
		]
		# fmt:on

	for path in search_paths:
		if path.exists():
			logging.debug(f"Lib {libname} found in {path!s}")
			return path

	# raise OSError(f"Can't find library '{libname}'. Searched at:\n" + "\n".join(str(p) for p in search_paths))
	return None


def get_phazor_path(pctl: PlayerCtl) -> Path:
	"""Locate the PHaZOR library in the specified priority order.

	Tries .so, .dll, .pyd, .dylib in that order and finally uses find_library as a fallback.

	:param pctl: PlayerCtl object containing installation details
	:return: Path to the library file
	:raises Exception: If no library is found
	"""
	# This is where compile-phazor.sh scripts place the dll
	base_path = Path(pctl.install_directory).parent.parent / "build"

	# Define the library name and extensions in priority order
	lib_name = "phazor-pw" if (not pctl.windows and not pctl.tauon.macos) and pctl.prefs.pipewire else "phazor"

	extensions = [".so", ".dll", ".pyd", ".dylib"]

	# Check explicitly for each file
	for ext in extensions:
		lib_path = base_path / f"lib{lib_name}{ext}"
		if lib_path.is_file():
			return lib_path

	# Fallback to find_library. Used if built using setuptools
	lib_path = find_library(lib_name)
	if lib_path:
		return Path(lib_path)

	raise Exception(f"Failed to load PHaZOR library ({lib_name})")


def phazor_exists(pctl: PlayerCtl) -> bool:
	"""Check for the existence of the PHaZOR library on the FS"""
	return get_phazor_path(pctl).exists()


def player4(tauon: Tauon) -> None:
	def scan_device() -> None:
		n = aud.scan_devices()
		devices = ["Default"]
		if n:
			for d in range(n):
				st = aud.get_device(d).decode()
				devices.append(st)
		prefs.phazor_devices = devices
		if prefs.phazor_device_selected not in devices:
			prefs.phazor_device_selected = devices[0]

	def pause_when_device_unavailable() -> None:
		pctl.pause_only()

	def calc_rg(track: TrackClass | None) -> float:
		if prefs.replay_gain == 0 and prefs.replay_preamp == 0:
			pctl.active_replaygain = 0
			return 0

		g = 0
		p = 1

		if track is not None:
			tg = track.misc.get("replaygain_track_gain")
			tp = track.misc.get("replaygain_track_peak")
			ag = track.misc.get("replaygain_album_gain")
			ap = track.misc.get("replaygain_album_peak")

			if prefs.replay_gain > 0:
				if prefs.replay_gain == 3 and tg is not None and ag is not None:
					gens = pctl.gen_codes.get(tauon.pl_to_id(pctl.active_playlist_playing))
					if pctl.random_mode or (gens and ("st" in gens or "rt" in gens or "r" in gens)):
						g = tg
						if tp is not None:
							p = tp
					else:
						g = ag
						if ap is not None:
							p = ap
				elif (prefs.replay_gain == 1 and tg is not None) or (
					prefs.replay_gain == 2 and ag is None and tg is not None
				):
					g = tg
					if tp is not None:
						p = tp
				elif ag is not None:
					g = ag
					if ap is not None:
						p = ap
				elif tg is not None and ag is None:
					g = tg
					if tp is not None:
						p = tp

		logging.debug("Detected ReplayGain")
		logging.debug("GAIN: " + str(g))
		logging.debug("PEAK: " + str(p))
		logging.debug("FINAL: " + str(min(10 ** ((g + prefs.replay_preamp) / 20), 1 / p)))

		if p == 0:
			logging.warning("Detected ReplayGain peak of 0")
			return 1
		pctl.active_replaygain = g
		return min(10 ** ((g + prefs.replay_preamp) / 20), 1 / p)

	def set_config(set_device: bool = False) -> None:
		aud.config_set_dev_buffer(prefs.device_buffer)
		aud.config_set_fade_duration(prefs.cross_fade_time)
		st = prefs.phazor_device_selected.encode()
		aud.config_set_dev_name(st)
		if set_device:
			aud.pause()
			aud.wait()
			aud.stop_out()
			aud.wait()
			aud.resume()
		if prefs.always_ffmpeg:
			aud.config_set_always_ffmpeg(1)
		if prefs.volume_power < 0 or prefs.volume_power > 10:
			prefs.volume_power = 2
		aud.config_set_volume_power(prefs.volume_power)
		aud.config_set_resample(prefs.avoid_resampling ^ True)
		apply_eq_settings()

	def normalise_eq_bands() -> list[float]:
		try:
			bands = list(getattr(prefs, "eq", []) or [])
		except TypeError:
			bands = []
		if len(bands) < 10:
			bands.extend([0.0] * (10 - len(bands)))
		elif len(bands) > 10:
			bands = bands[:10]
		out: list[float] = []
		for v in bands:
			try:
				fv = float(v)
			except (TypeError, ValueError):
				fv = 0.0
			out.append(float(max(min(fv, 12.0), -12.0)))
		bands = out
		prefs.eq = bands
		return bands

	def apply_eq_settings() -> None:
		if not hasattr(aud, "eq_set_enable") or not hasattr(aud, "eq_set_band"):
			return
		aud.eq_set_enable(1 if prefs.use_eq else 0)
		for i, gain in enumerate(normalise_eq_bands()):
			aud.eq_set_band(i, ctypes.c_float(gain))

	def run_vis() -> None:
		if gui.turbo:  # and pctl.playing_time > 0.5:
			if gui.vis == 2:
				p_spec: list[float] = []
				aud.get_spectrum(24, bins1)
				bias = 1
				for b in list(bins1):
					p_spec.append(int(b * 1.7 * bias))
					bias += 0.04
				gui.spec = p_spec
				gui.level_update = True
				if pctl.playing_time > 0.5 and (pctl.playing_state in (PlayingState.PLAYING, PlayingState.URL_STREAM)):
					gui.update_spec = 1
			elif gui.vis == 4:
				p_spec: list[float] = []
				aud.get_spectrum(45, bins2)
				bias = 1
				for b in list(bins2):
					p_spec.append(int(b * 2.0 * bias))
					bias += 0.01
				gui.spec4_array = p_spec
				gui.level_update = True
				if pctl.playing_time > 0.5 and (pctl.playing_state in (PlayingState.PLAYING, PlayingState.URL_STREAM)):
					gui.update_spec = 1

	p_sync_timer = Timer()

	def track(end: bool = True) -> None:
		run_vis()

		if end and loaded_track and loaded_track.is_network and pctl.playing_time < 7 and aud.get_result() == 2:
			logging.info("STALL, RETRY")
			time.sleep(0.5)
			pctl.playerCommandReady = True
			pctl.playerCommand = "open"

		add_time = player_timer.hit()
		if add_time > 2:
			add_time = 2
		elif add_time < 0:
			add_time = 0

		pctl.total_playtime += add_time

		# Wait / speed up, if we are out of sync
		if p_sync_timer.get() > 1:
			real_position = aud.get_position_ms() / 1000
			if real_position and pctl.playing_time and real_position != pctl.last_real_position:
				diff = abs(real_position - pctl.playing_time)
				if 5 > diff > 0.11:  # in a CUE file real will be different that playing time
					# This assumes the first track in a CUE is > 5s
					if real_position < pctl.playing_time:
						add_time -= 5
						add_time = max(add_time, 0)
						p_sync_timer.force_set(2)  # wait for real to catch up again next clock
					else:
						add_time += 0.1
				pctl.last_real_position = real_position  # we still want to move on if playback stalled
			p_sync_timer.set()

		pctl.playing_time += add_time
		pctl.decode_time = pctl.playing_time

		if pctl.playing_time < 3 and pctl.a_time < 3:
			pctl.a_time = pctl.playing_time
		else:
			pctl.a_time += add_time

		tauon.lfm_scrobbler.update(add_time)

		if len(pctl.track_queue) > 0 and 2 > add_time > 0:
			tauon.star_store.add(pctl.track_queue[pctl.queue_step], add_time)
		if end and pctl.playing_time > 1:
			pctl.test_progress()

	def chrome_start(track_id: int, enqueue: bool = False, t: int = 0) -> None:
		track = pctl.get_track(track_id)
		# if track.is_cue:
		# 	logging.error("CUE cast not supported")
		# 	return
		if track.is_network:
			if track.file_ext == "SPTY":
				logging.error("Spotify cast not supported")
				return
			network_url, params = pctl.get_url(track)
			if params:
				req = PreparedRequest()
				req.prepare_url(network_url, params)
				network_url = req.url

			tauon.chrome.start(track.index, enqueue=enqueue, url=network_url, t=t)
		else:
			tauon.chrome.start(track.index, enqueue=enqueue, t=t)

	def start_librespot() -> None:
		logging.info("SPP: Start librespot command received. Set Phazor input raw mode")
		aud.start(b"RAW FEED", 0, 0, ctypes.c_float(calc_rg(None)))
		spotc.go(force=True)
		if not spotc.running:
			shooter(spotc.worker)

	# fmt:off
	gui   = tauon.gui
	pctl  = tauon.pctl
	prefs = tauon.prefs
	# fmt:on
	logging.debug("Starting PHAzOR backend…")

	player_timer = Timer()
	loaded_track = None
	fade_time = 400

	aud = tauon.aud

	aud.config_set_dev_name(prefs.phazor_device_selected.encode())

	aud.init()

	aud.get_device.restype = ctypes.c_char_p

	aud.feed_raw.argtypes = (ctypes.c_int, ctypes.c_char_p)
	aud.feed_raw.restype = None
	aud.set_volume(int(pctl.player_volume))

	try:
		aud.eq_set_enable.argtypes = (ctypes.c_int,)
		aud.eq_set_enable.restype = None
		aud.eq_set_band.argtypes = (ctypes.c_int, ctypes.c_float)
		aud.eq_set_band.restype = None
	except AttributeError:
		logging.warning("PHAzOR build does not expose EQ controls")

	bins1 = (ctypes.c_float * 24)()
	bins2 = (ctypes.c_float * 45)()

	aud.get_level_peak_l.restype = ctypes.c_float
	aud.get_level_peak_r.restype = ctypes.c_float

	active_timer = Timer()

	scan_device()
	spotc = tauon.spotc
	ff_run = FFRun(tauon)

	if sys.platform == "win32":
		FUNCTYPE = WINFUNCTYPE
	else:
		FUNCTYPE = CFUNCTYPE
	start_callback = FUNCTYPE(c_int, c_char_p, c_int, c_int)(ff_run.start)
	read_callback = FUNCTYPE(c_int, c_void_p, c_int)(ff_run.read)
	close_callback = FUNCTYPE(c_void_p)(ff_run.close)
	device_unavailable_callback = FUNCTYPE(c_void_p)(pause_when_device_unavailable)
	aud.set_callbacks(start_callback, read_callback, close_callback, device_unavailable_callback)

	cachement = tauon.cachement

	# aud.config_set_samplerate(prefs.samplerate)
	aud.config_set_resample_quality(prefs.resample)

	set_config()

	stall_timer = Timer()
	wall_timer = Timer()

	chrome_update = 0
	chrome_cool_timer = Timer()
	chrome_mode = False

	while True:
		# logging.error(aud.print_status())
		time.sleep(0.016)
		if tauon.player4_state == PlayerState.PAUSED:
			time.sleep(0.05)
		if (
			tauon.player4_state != PlayerState.STOPPED
			or tauon.spot_ctl.playing
			or tauon.spot_ctl.coasting
			or tauon.chrome_mode
		):
			active_timer.set()
		if active_timer.get() > 7:
			aud.stop()
			aud.phazor_shutdown()
			spotc.soft_end()
			break

		# Level meter
		if (tauon.player4_state in (PlayerState.PLAYING, PlayerState.URL_STREAM)) and gui.vis == 1:
			amp = aud.get_level_peak_l()
			l = amp * 12
			amp = aud.get_level_peak_r()
			r = amp * 12

			tauon.level_train.append((0, l, r))
			gui.level_update = True

		if chrome_mode:
			if tauon.chrome is None:
				logging.critical("This should not happen, tauon.chrome was None")
				return
			if pctl.playerCommandReady:
				command = pctl.playerCommand
				# logging.info(command)
				subcommand = pctl.playerSubCommand
				pctl.playerSubCommand = ""
				pctl.playerCommandReady = False

				if command == "endchrome":
					chrome_mode = False
					tauon.player4_state = PlayerState.STOPPED
					pctl.playing_time = 0
					pctl.decode_time = 0
					pctl.stop()
					continue
				if command == "open":
					target_object = pctl.target_object
					if tauon.player4_state == PlayerState.PLAYING:
						t, pid, s, d = tauon.chrome.update()
						# logging.info((t, d))
						# logging.info(d - t)

						if d and t and 1 < d - t < 5:
							# logging.info("Enqueue next chromecast")
							chrome_start(target_object.index, enqueue=True, t=pctl.start_time_target)
							chrome_cool_timer.set()
							time.sleep(d - t)
							if pctl.commit:
								pctl.advance(quiet=True, end=True)
								pctl.commit = None
							continue

					chrome_start(target_object.index, t=pctl.start_time_target)
					chrome_cool_timer.set()
					if pctl.commit:
						pctl.advance(quiet=True, end=True)
						pctl.commit = None
					tauon.player4_state = PlayerState.PLAYING
				if command == "pauseon":
					tauon.chrome.pause()
					tauon.player4_state = PlayerState.PAUSED
				if command == "pauseoff":
					tauon.chrome.play()
					tauon.player4_state = PlayerState.PLAYING
				if command == "volume":
					tauon.chrome.volume(round(pctl.player_volume / 100, 3))
					tauon.player4_state = PlayerState.PLAYING
				if command == "seek":
					tauon.chrome.seek(float(round(pctl.new_time + pctl.start_time_target, 2)))
					chrome_cool_timer.set()
					pctl.playing_time = pctl.new_time
					pctl.decode_time = pctl.playing_time
				if command == "seteq":
					apply_eq_settings()
				if command == "stop":
					tauon.player4_state = PlayerState.STOPPED
					tauon.chrome.stop()

			if tauon.player4_state == PlayerState.PLAYING:
				if chrome_update > 0.8 and chrome_cool_timer.get() > 2.5:
					t, pid, s, d = tauon.chrome.update()
					pctl.playing_time = t - pctl.start_time_target
					pctl.decode_time = t - pctl.start_time_target
					player_timer.hit()
					chrome_update = 0

				add_time = player_timer.hit()
				add_time = min(add_time, 2)
				chrome_update += add_time
				pctl.a_time += add_time
				tauon.lfm_scrobbler.update(add_time)
				pctl.total_playtime += add_time
				if len(pctl.track_queue) > 0 and 2 > add_time > 0:
					tauon.star_store.add(pctl.track_queue[pctl.queue_step], add_time)

				pctl.test_progress()

			time.sleep(0.1)
			continue

		# Command processing
		if pctl.playerCommandReady:
			command = pctl.playerCommand
			subcommand = pctl.playerSubCommand
			pctl.playerSubCommand = ""
			pctl.playerCommandReady = False
			# logging.info(command)

			if command == "spotcon":
				# aud.stop()
				logging.info("SPP: Start librespot command received. Set Phazor input raw mode")
				start_librespot()
				tauon.player4_state = PlayerState.SPOTIFY_MODE
				# time.sleep(20)
				# tauon.spot_ctl.preparing_spotify = False

			if command == "startchrome":
				aud.stop()
				if tauon.player4_state == PlayerState.PLAYING:
					chrome_start(loaded_track.index, t=pctl.playing_time)
				chrome_mode = True

			if command == "reload":
				set_config()
			if command == "set-device":
				set_config(set_device=True)

			if command == "url":
				pctl.download_time = 0
				w = 0
				if not tauon.radiobox.run_proxy:
					aud.start(pctl.url.encode(), 0, 0, ctypes.c_float(calc_rg(None)))
					tauon.player4_state = PlayerState.URL_STREAM
					player_timer.hit()
				else:
					while len(tauon.stream_proxy.chunks) < 200:
						time.sleep(0.1)
						w += 1
						if w > 100:
							logging.info("Taking too long!")
							tauon.stream_proxy.stop()
							pctl.playerCommand = "stop"
							pctl.playerCommandReady = True
							break
					else:
						aud.config_set_feed_samplerate(prefs.samplerate)
						aud.start(b"RAW FEED", 0, 0, ctypes.c_float(calc_rg(None)))
						tauon.player4_state = PlayerState.URL_STREAM
						player_timer.hit()

			if command == "open":
				if tauon.player4_state == PlayerState.PAUSED:
					aud.set_volume(int(pctl.player_volume))

				stall_timer.set()
				pctl.download_time = 0
				target_object = pctl.target_object
				if not target_object:
					logging.exception("This shouldn't happen, target_object was None")
					continue
				target_path = target_object.fullpath
				subtrack = target_object.subtrack
				aud.set_subtrack(subtrack)

				logging.info(
					f"Open - requested start was {(pctl.start_time_target + pctl.jump_time)} ({pctl.start_time_target})"
				)
				try:
					logging.info(f"Extension: {target_path.split('.')[-1]}")
				except Exception:
					logging.exception("Failed to get extension - maybe file name does not have any dots?")

				if (tauon.spot_ctl.playing or tauon.spot_ctl.coasting) and target_object.file_ext != "SPTY":
					tauon.spot_ctl.control("stop")

				if target_object.is_network:
					if target_object.file_ext == "SPTY":
						tauon.level_train.clear()
						if target_object.found is False:
							pctl.playing_state = PlayingState.STOPPED
							pctl.jump_time = 0.0
							pctl.advance(inplace=True, play=True)
							continue
						if tauon.player4_state not in (PlayerState.STOPPED, PlayerState.SPOTIFY_MODE):
							aud.stop()
						if tauon.player4_state != PlayerState.SPOTIFY_MODE:
							tauon.player4_state = PlayerState.STOPPED

						try:
							f = False
							if spotc.running and tauon.player4_state != PlayerState.SPOTIFY_MODE:
								aud.start(b"RAW FEED", 0, 0, ctypes.c_float(calc_rg(None)))
								tauon.player4_state = PlayerState.SPOTIFY_MODE
							if prefs.launch_spotify_local and not spotc.running:
								aud.start(b"RAW FEED", 0, 0, ctypes.c_float(calc_rg(None)))
								tauon.player4_state = PlayerState.SPOTIFY_MODE
								if not tauon.librespot_p:
									tauon.spot_ctl.preparing_spotify = True
									f = True
								spotc.go()
								if not spotc.running:
									shooter(spotc.worker)

							tauon.spot_ctl.play_target(
								target_object.url_key, force_new_device=f, start_callback=start_librespot
							)
							# tauon.spot_ctl.preparing_spotify = False
							if pctl.playerCommand == "spotcon":  # such spaghetti code
								tauon.player4_state = PlayerState.SPOTIFY_MODE
								pctl.playerCommand = ""
								pctl.playerCommandReady = False

						except Exception:
							logging.exception("Failed to start Spotify track")
							# tauon.spot_ctl.preparing_spotify = False
							pctl.playerCommand = "stop"
							pctl.playerCommandReady = True
						continue

					spotc.running = False
					# tauon.spot_ctl.preparing_spotify = False
					pctl.spot_playing = False
					timer = Timer()
					timer.set()
					while True:
						status, path = cachement.get_file(target_object)

						if status in (0, 2):
							break
						if timer.get() > 0.25 and gui.buffering is False:
							gui.buffering_text = ""
							pctl.buffering_percent = 0
							gui.buffering = True
							gui.update += 1
							tauon.wake()
						if cachement.ready == target_object and pctl.start_time_target + pctl.jump_time == 0.0:
							break
						time.sleep(0.05)
						# logging.info(status)

					gui.buffering = False
					gui.update += 1
					tauon.wake()

					if status == 2:
						logging.info("Could not locate resource")
						target_object.found = False
						pctl.playing_state = PlayingState.STOPPED
						pctl.jump_time = 0.0
						# pctl.advance(inplace=True, play=True)
						continue
					target_path = path

				elif prefs.precache:
					timer = Timer()
					timer.set()
					while True:
						status, path = cachement.get_file(target_object)
						if status in (0, 2):
							break
						if timer.get() > 0.25 and gui.buffering is False:
							gui.buffering_text = ""
							pctl.buffering_percent = 0
							gui.buffering = True
							gui.update += 1
							tauon.wake()

						time.sleep(0.05)

					gui.buffering = False
					gui.update += 1
					tauon.wake()

					if status == 2:
						target_object.found = False
						pctl.playing_state = PlayingState.STOPPED
						pctl.jump_time = 0.0
						pctl.advance(inplace=True, play=True)
						continue
					target_path = path

				if not os.path.isfile(target_path):
					target_object.found = False
					if not target_object.is_network:
						pctl.playing_state = PlayingState.STOPPED
						pctl.jump_time = 0.0
						pctl.advance(inplace=True, play=True)
					continue
				if not target_object.found:
					pctl.reset_missing_flags()

				spotc.running = False
				length: float = 0
				remain: float = 0
				position: float = 0

				if target_path and target_object and target_object.length == 0 and not target_object.is_cue:
					logging.info("Track has duration of 0, scanning file")
					temp = tauon.TrackClass()
					temp.fullpath = target_path
					tauon.tag_scan(temp)
					target_object.length = temp.length
					if pctl.playing_object() is target_object:
						pctl.playing_length = target_object.length
					del temp

				if (
					tauon.player4_state == PlayerState.PLAYING
					and not pctl.start_time_target
					and not pctl.jump_time
					and loaded_track
				):
					length = aud.get_length_ms() / 1000
					position = aud.get_position_ms() / 1000
					remain = length - position

					# TODO(Martin): The GUI logger does not support multiline
					# logging.info(f"{loaded_track.title} -> {target_object.title}\nlength: {length!s}\nposition: {position!s}\nWe are {remain!s} from end")
					logging.info(f"{loaded_track.title} -> {target_object.title}")
					logging.info(f" --- length: {length!s}")
					logging.info(f" --- position: {position!s}")
					logging.info(f" --- We are {remain!s} from end")

					if loaded_track.is_network or length == 0:
						logging.warning("Phazor did not respond with a duration")
						length = loaded_track.length
						remain = length - position

				fade = 0
				error = False
				if (
					tauon.player4_state == PlayerState.PLAYING
					and length
					and position
					and not pctl.start_time_target
					and not pctl.jump_time
					and loaded_track
					and 0 < remain < 5.5
					and not loaded_track.is_cue
					and subcommand != "now"
				):
					logging.info("Transition gapless")

					r_timer = Timer()
					r_timer.set()

					if loaded_track and loaded_track.file_ext.lower() in tauon.formats.GME:
						# GME formats dont have a physical end so we don't do gapless
						while r_timer.get() <= remain - prefs.device_buffer / 1000:
							if pctl.commit:
								track(end=False)
							time.sleep(0.016)
						aud.stop()

					aud.next(
						target_path.encode(),
						int((pctl.start_time_target + pctl.jump_time) * 1000),
						ctypes.c_float(calc_rg(target_object)),
					)

					cont = False
					while r_timer.get() <= remain - prefs.device_buffer / 1000:
						if pctl.commit:
							track(end=False)
						time.sleep(0.016)
						if pctl.playerCommandReady and pctl.playerCommand in ("open", "stop"):
							logging.info("JANK")
							pctl.commit = None
							break
						if pctl.playerCommandReady and pctl.playerCommand == "seek":
							logging.info("Seek revert gapless")
							pctl.commit = None
							pctl.jump(loaded_track.index, jump=True)
							pctl.jump_time = pctl.new_time
							pctl.playing_time = pctl.new_time
							pctl.decode_time = pctl.new_time
							cont = True
							break

					if cont:
						continue
					if pctl.commit is not None:
						pctl.playing_time = 0
						pctl.decode_time = 0
						match = pctl.commit
						if match == loaded_track.index and subcommand == "repeat":
							pctl.update_change()
						else:
							pctl.advance(quiet=True, end=True)
						pt = pctl.playing_object()
						if pt and pt.index != match:
							logging.info("MISSFIRE")
							pctl.play_target()
							continue
						if pctl.playerCommandReady and pctl.playerCommand == "open":
							pctl.playerCommandReady = False
							pctl.playerCommand = ""

					loaded_track = target_object
					pctl.playing_time = pctl.jump_time
				else:
					if pctl.commit and subcommand != "repeat":
						pctl.advance(quiet=True, end=True)
						pctl.commit = None
						continue

					if tauon.player4_state == PlayerState.PLAYING and prefs.use_jump_crossfade:
						fade = 1

					logging.info("Transition jump")
					aud.start(
						target_path.encode(errors="surrogateescape"),
						int((pctl.start_time_target + pctl.jump_time) * 1000),
						fade,
						ctypes.c_float(calc_rg(target_object)),
					)
					loaded_track = target_object
					pctl.playing_time = pctl.jump_time
					if pctl.jump_time:
						while aud.get_result() == 0:
							time.sleep(0.016)
							run_vis()
						aud.set_position_ms(int(pctl.jump_time * 1000))

					# Restart track is failed to load (for some network tracks) (broken with gapless loading)
					while True:
						r = aud.get_result()
						if r == 1:
							break
						if r == 2:
							if loaded_track.is_network:
								pctl.buffering_percent = 0
								gui.buffering = True
								gui.buffering_text = ""

								# while dm.request(loaded_track, whole=True) == "wait":
								# 	time.sleep(0.05)
								# 	if pctl.playerCommandReady:
								# 		break
								logging.info("Retry start file")
								aud.start(
									target_path.encode(),
									int((pctl.start_time_target + pctl.jump_time) * 1000),
									fade,
									ctypes.c_float(calc_rg(target_object)),
								)
								gui.buffering = False
								player_timer.set()
								break
							aud.stop()
							if not gui.message_box:
								tauon.show_message(_("Error loading track"), mode="warning")
							error = True
							break
						time.sleep(0.016)
						run_vis()

					tauon.player4_state = PlayerState.PLAYING
					if error:
						tauon.player4_state = PlayerState.STOPPED

				player_timer.set()
				pctl.jump_time = 0.0
				if loaded_track.length == 0 or loaded_track.file_ext.lower() in tauon.formats.MOD:
					i = 0
					t = 0
					while t == 0:
						time.sleep(0.3)
						t = aud.get_length_ms() / 1000
						i += 1
						if i > 9:
							break
					loaded_track.length = t
					if loaded_track.length != 0:
						pctl.playing_length = loaded_track.length
						gui.pl_update += 1

				pctl.commit = None
				stall_timer.set()
				wall_timer.force_set(3)

			if command == "seek":
				if tauon.spot_ctl.coasting or tauon.spot_ctl.playing:
					tauon.spot_ctl.control("seek", int(pctl.new_time * 1000))
					pctl.playing_time = pctl.new_time
				elif tauon.player4_state != PlayerState.STOPPED:
					if loaded_track.is_network:  # and loaded_track.fullpath.endswith(".ogg"):
						timer = Timer()
						timer.set()
						i = 0
						while True:
							status, path = cachement.get_file(loaded_track)
							if status == 1:
								per = (pctl.new_time / loaded_track.length) * 100
								if per < 1:
									break
								if pctl.buffering_percent - per > 5:
									break
							if status in (0, 2):
								break
							if timer.get() > 0.25 and gui.buffering is False:
								gui.buffering_text = ""
								pctl.buffering_percent = 0
								gui.buffering = True
								gui.update += 1
								tauon.wake()
							if i * 0.05 > 2:
								aud.pause()
							if pctl.playerCommandReady:
								break

							time.sleep(0.05)
							i += 1

						gui.buffering = False
						gui.update += 1
						tauon.wake()
						if pctl.playerCommandReady:
							continue

						if status == 2:
							loaded_track.found = False
							pctl.playing_state = PlayingState.STOPPED
							pctl.jump_time = 0.0
							pctl.stop()
							continue

						# The vorbis decoder doesn't like appended files
						aud.start(
							path.encode(),
							int(pctl.new_time + pctl.start_time_target) * 1000,
							0,
							ctypes.c_float(calc_rg(loaded_track)),
						)
						while aud.get_result() == 0:
							time.sleep(0.01)
					else:
						aud.seek(int((pctl.new_time + pctl.start_time_target) * 1000), prefs.pa_fast_seek)

					pctl.playing_time = pctl.new_time

				pctl.decode_time = pctl.playing_time
				wall_timer.set()

			if command == "volume":
				if (tauon.spot_ctl.coasting or tauon.spot_ctl.playing) and not spotc.running:
					tauon.spot_ctl.control("volume", int(pctl.player_volume))
				else:
					aud.ramp_volume(int(pctl.player_volume), 750)

			if command == "seteq":
				apply_eq_settings()

			if command == "runstop":
				length = aud.get_length_ms() / 1000
				position = aud.get_position_ms() / 1000
				remain = length - position
				# logging.info("length: " + str(length))
				# logging.info("position: " + str(position))
				# logging.info("We are %s from end" % str(remain))
				time.sleep(3)
				command = "stop"

			if command == "stop":
				if prefs.use_pause_fade and tauon.player4_state != PlayerState.URL_STREAM:
					if pctl.player_volume > 5:
						speed = fade_time / (int(pctl.player_volume) / 100)
					else:
						speed = fade_time / (5 / 100)

					aud.ramp_volume(0, int(speed))
					time.sleep((fade_time + 100) / 1000)

				tauon.player4_state = PlayerState.STOPPED
				pctl.playing_time = 0
				aud.stop()
				spotc.running = False
				time.sleep(0.1)
				aud.set_volume(int(pctl.player_volume))

				if subcommand == "return":
					pctl.playerSubCommand = "stopped"
					# pctl.playerCommandReady = True

			if command == "pauseon":
				if prefs.use_pause_fade:
					if pctl.player_volume > 5:
						speed = fade_time / (int(pctl.player_volume) / 100)
					else:
						speed = fade_time / (5 / 100)

					aud.ramp_volume(0, int(speed))
					time.sleep((fade_time + 100) / 1000)
				aud.pause()
				tauon.player4_state = PlayerState.PAUSED

			if command == "pauseoff":
				if tauon.player4_state == PlayerState.STOPPED:
					t = pctl.playing_time
					pctl.play_target()
					pctl.jump_time = t
					pctl.playing_time = t
					pctl.decode_time = t
				else:
					if prefs.use_pause_fade:
						if pctl.player_volume > 5:
							speed = fade_time / (int(pctl.player_volume) / 100)
						else:
							speed = fade_time / (5 / 100)

						aud.ramp_volume(int(pctl.player_volume), int(speed))
					aud.resume()
					player_timer.set()
					stall_timer.set()
					tauon.player4_state = PlayerState.PLAYING

			if command == "unload":
				if prefs.use_pause_fade:
					if pctl.player_volume > 5:
						speed = fade_time / (int(pctl.player_volume) / 100)
					else:
						speed = fade_time / (5 / 100)

					aud.ramp_volume(0, int(speed))
					time.sleep((fade_time + 100) / 1000)

				spotc.end()
				aud.stop()
				aud.phazor_shutdown()

				if cachement.audio_cache.exists():
					shutil.rmtree(cachement.audio_cache)

				pctl.playerCommand = "done"
				pctl.playerCommandReady = True
				break
		else:
			pctl.spot_test_progress()

			if tauon.player4_state == PlayerState.SPOTIFY_MODE:
				run_vis()

			if tauon.player4_state == PlayerState.URL_STREAM:
				pctl.radio_progress()
				run_vis()

				add_time = player_timer.hit()
				pctl.playing_time += add_time
				pctl.decode_time = pctl.playing_time

				buffering = aud.is_buffering()
				if gui.buffering != buffering:
					gui.buffering = buffering
					gui.update += 1

			if tauon.player4_state == PlayerState.PLAYING:
				track()



"""Tauon Music Box - Tidal streaming module"""

from __future__ import annotations

import json
import logging
import os
import webbrowser
from datetime import datetime, timezone
from typing import TYPE_CHECKING

allow_tidal = False
try:
	from tidalapi.media import Quality, Track
	from tidalapi.session import Session
	from tidalapi.user import Favorites

	allow_tidal = True
except ModuleNotFoundError:
	logging.warning("Unable to import tidalapi, Tidal support will be disabled.")
except Exception:
	logging.exception("Tidalapi not found")

if TYPE_CHECKING:
	from tauon.t_modules.t_main import Tauon, TrackClass


class Tidal:
	def __init__(self, tauon: Tauon) -> None:
		self.tauon = tauon
		self.session: Session | None = None
		self.save_path = str(tauon.cache_directory / "tidal.json")
		self.login_stage: int = 0
		self.import_cache: dict[str, TrackClass] = {}

	def logout(self) -> None:
		self.session = None
		self.login_stage = 0
		if os.path.isfile(self.save_path):
			os.remove(self.save_path)

	def login1(self) -> None:
		if not allow_tidal:
			self.tauon.show_message(_("Tidalapi package not loaded"))
			return
		logging.info("LOGIN 1")
		session = Session()
		# session.config.pkce_uri_redirect = f"http://localhost:7811/tidalredir"
		login_url = session.pkce_login_url()
		webbrowser.open(login_url, new=2, autoraise=True)
		self.session = session
		self.login_stage = 2

	def login2(self, url: str) -> None:
		logging.info("LOGIN 2")
		if not allow_tidal:
			return
		auth_token = self.session.pkce_get_auth_token(url)
		self.session.process_auth_token(auth_token, is_pkce_token=True)
		logging.info("login2 done")
		self.save_session()
		self.login_stage = 0
		# https://tidal.com/android/login/auth?code=eyVyPty.WiGdw-abQ&state=na&appMode=android

	def try_load(self) -> None:
		if not allow_tidal:
			self.tauon.show_message(_("Tidalapi package not loaded"))
			return
		if not self.session and os.path.isfile(self.save_path):
			with open(self.save_path) as f:
				session_data = json.load(f)
			session = Session()

			expiry_time = None
			if session_data["expiry_time"]:
				expiry_time = datetime.fromisoformat(session_data["expiry_time"])
				# Ensure the datetime is timezone-aware
				if expiry_time.tzinfo is None:
					expiry_time = expiry_time.replace(tzinfo=timezone.utc)

			success = session.load_oauth_session(
				session_data["token_type"],
				session_data["access_token"],
				session_data["refresh_token"],
				expiry_time,
				session_data["is_pkce"],
			)
			if success:
				self.session = session
				self.save_session()
				logging.info("loaded TIDAL login")
			else:
				logging.error("Failed to load TIDAL login")
				self.tauon.show_message(_("Failed to load TIDAL login, please try log in again"))
				os.remove(self.save_path)

	def save_session(self) -> None:
		if self.session:
			session = self.session
			session_data = {
				"token_type": session.token_type,
				"access_token": session.access_token,
				"refresh_token": session.refresh_token,
				"expiry_time": session.expiry_time.isoformat() if session.expiry_time else None,
				"is_pkce": session.is_pkce,
			}
			with open(self.save_path, "w", encoding="utf-8") as f:
				json.dump(session_data, f)

	def resolve_stream(self, tr: TrackClass) -> list[str] | None:
		track_id = tr.url_key
		self.try_load()
		if not self.session:
			logging.info("Tidal: not logged in")
			return None

		if self.tauon.prefs.tidal_quality == 2:
			logging.info("TRY 24 bit flac")
			self.session.audio_quality = Quality.hi_res_lossless
			try:
				result = self.resolve_stream2(tr)
				if result:
					return result
			except Exception:
				logging.exception("Failed to resolve stream")

		if self.tauon.prefs.tidal_quality == 1:
			logging.info("TRY 16bit flac")
			self.session.audio_quality = Quality.high_lossless
			try:
				result = self.resolve_stream2(tr)
				if result:
					return result
			except Exception:
				logging.exception("Failed to resolve stream")

		logging.info("TRY LOW")
		self.session.audio_quality = Quality.low_320k
		try:
			result = self.resolve_stream2(tr)
			if result:
				return result
		except Exception:
			logging.exception("Failed to resolve stream")

	def resolve_stream2(self, tr: TrackClass) -> list[str] | None:
		track_id = tr.url_key
		track = self.session.track(track_id)
		logging.info(f"{track.id}: '{track.name}' by '{track.artist.name}'")
		stream = track.get_stream()
		logging.info(f"MimeType:{stream.manifest_mime_type}")

		manifest = stream.get_stream_manifest()
		audio_resolution = stream.get_audio_resolution()

		logging.info(
			f"track:{track.id}, (quality:{stream.audio_quality}, codec:{manifest.get_codecs()}, {audio_resolution[0]}bit/{audio_resolution[1]}Hz)"
		)
		tr.misc["container"] = manifest.get_codecs().upper()
		tr.samplerate = str(audio_resolution[1])
		tr.bit_depth = audio_resolution[0]
		if stream.is_mpd:
			logging.info("MPD!")
			return manifest.get_urls()
		if stream.is_bts:
			logging.info("BTS!")
			return manifest.get_urls()

		return None

	def build_cache(self) -> None:
		for id, nt in self.tauon.pctl.master_library.items():
			if nt.url_key and nt.file_ext == "TIDAL":
				self.import_cache[nt.url_key] = nt

	def new_track(self, track: Track) -> TrackClass:
		new = False
		track_id = str(track.id)
		nt = self.import_cache.get(track_id)

		if not nt:
			new = True
			nt = self.tauon.TrackClass()
			nt.index = self.tauon.pctl.master_count

		nt.is_network = True
		nt.file_ext = "TIDAL"
		nt.url_key = track_id

		nt.track_number = str(track.track_num)
		nt.title = track.name
		nt.artist = track.artist.name
		nt.album = track.album.name
		nt.length = track.duration
		nt.album_artist = track.album.artist.name
		nt.misc["tidal_album"] = track.album.id

		parent = (nt.album_artist + " - " + nt.album).strip("- ")
		nt.parent_folder_path = parent
		nt.parent_folder_name = parent

		nt.art_url_key = ""
		if track.album.cover:
			nt.art_url_key = track.album.image(dimensions=1280)

		if new:
			self.tauon.pctl.master_count += 1
			self.tauon.pctl.master_library[nt.index] = nt

		return nt

	def track(self, id: str) -> list | None:
		self.try_load()
		if not self.session:
			return []
		self.build_cache()

		t = self.session.track(id)
		nt = self.new_track(t)
		self.tauon.pctl.multi_playlist[self.tauon.pctl.active_playlist_viewing].playlist_ids.append(nt.index)
		self.tauon.gui.pl_update += 14
		return None

	def fav_albums(self, return_list: bool = False) -> list[TrackClass] | None:
		self.try_load()
		if not self.session:
			return []
		self.build_cache()

		try:
			f = Favorites(self.session, self.session.user.id)
		except Exception:
			logging.exception("Error getting tidal user favorites")
			return []

		playlist = []
		for album in f.albums():
			for track in album.tracks():
				nt = self.new_track(track)
				playlist.append(nt.index)

		if return_list:
			return playlist

		self.tauon.pctl.multi_playlist.append(self.tauon.pl_gen(title="TIDAL Albums", playlist_ids=playlist))
		self.tauon.pctl.gen_codes[self.tauon.pl_to_id(len(self.tauon.pctl.multi_playlist) - 1)] = "tfa"
		self.tauon.show_message("Playlist load complete", mode="done")
		return None

	def fav_tracks(self, return_list: bool = False) -> list[TrackClass] | None:
		self.try_load()
		if not self.session:
			return []
		self.build_cache()

		try:
			f = Favorites(self.session, self.session.user.id)
		except Exception:
			logging.exception("Error getting tidal user favorites")
			return []

		playlist = []
		for track in f.tracks():
			nt = self.new_track(track)
			playlist.append(nt.index)

		if return_list:
			return playlist

		self.tauon.pctl.multi_playlist.append(self.tauon.pl_gen(title="TIDAL Tracks", playlist_ids=playlist))
		self.tauon.pctl.gen_codes[self.tauon.pl_to_id(len(self.tauon.pctl.multi_playlist) - 1)] = "tft"
		self.tauon.show_message("Playlist load complete", mode="done")
		return None

	def playlist(self, id: str, return_list: bool = False) -> list[TrackClass] | None:
		self.try_load()
		if not self.session:
			return []
		self.build_cache()

		try:
			p = self.session.playlist(id)
		except Exception:
			logging.exception("Error getting tidal playlist")
			return []

		playlist = []
		for track in p.tracks():
			nt = self.new_track(track)
			playlist.append(nt.index)

		if return_list:
			return playlist

		self.tauon.pctl.multi_playlist.append(self.tauon.pl_gen(title=p.name, playlist_ids=playlist))
		self.tauon.pctl.gen_codes[self.tauon.pl_to_id(len(self.tauon.pctl.multi_playlist) - 1)] = f'tpl"{id}"'
		return None

	def mix(self, id: str, return_list: bool = False) -> list[TrackClass] | None:
		self.try_load()
		if not self.session:
			return []
		self.build_cache()

		try:
			p = self.session.mix(id)
		except Exception:
			logging.exception("Error getting tidal mix")
			return []

		playlist = []
		for item in p.items():
			if item.type in ("track", None):  # tracks are None, api bug?
				nt = self.new_track(item)
				playlist.append(nt.index)

		if return_list:
			return playlist

		self.tauon.pctl.multi_playlist.append(self.tauon.pl_gen(title=p.title, playlist_ids=playlist))
		self.tauon.pctl.gen_codes[self.tauon.pl_to_id(len(self.tauon.pctl.multi_playlist) - 1)] = f'tmix"{id}"'
		return None

	def artist(self, id: str, return_list: bool = False) -> list[TrackClass] | None:
		self.try_load()
		if not self.session:
			return []
		self.build_cache()

		try:
			a = self.session.artist(id)
		except Exception:
			logging.exception("Error getting tidal artist")
			return []

		playlist = []

		for album in a.get_albums():
			for track in album.tracks():
				nt = self.new_track(track)
				playlist.append(nt.index)

		if return_list:
			return playlist

		self.tauon.pctl.multi_playlist.append(self.tauon.pl_gen(title=a.name, playlist_ids=playlist))
		self.tauon.pctl.gen_codes[self.tauon.pl_to_id(len(self.tauon.pctl.multi_playlist) - 1)] = f'tar"{id}"'
		return None

	def append_album(self, id: str) -> None:
		self.try_load()
		if not self.session:
			return

		self.build_cache()

		album = self.session.album(id)
		tracks = album.tracks()

		for track in tracks:
			logging.info(f"{track.id}: '{track.name}' by '{track.artist.name}'")

			nt = self.new_track(track)
			self.tauon.pctl.multi_playlist[self.tauon.pctl.active_playlist_viewing].playlist_ids.append(nt.index)
			self.tauon.gui.pl_update += 1

	# def test(self):
	# 	 logging.info("Test Tidal")
	# 	 self.try_load()
	# 	 if not self.session:
	# 		 logging.info("Tidal: not logged in")
	# 		 return
	#
	# 	 session = self.session
	# 	 # Override the required playback quality, if necessary
	# 	 # Note: Set the quality according to your subscription.
	# 	 # Low: Quality.low_96k          (m4a 96k)
	# 	 # Normal: Quality.low_320k      (m4a 320k)
	# 	 # HiFi: Quality.high_lossless   (FLAC)
	# 	 # HiFi+ Quality.hi_res          (FLAC MQA)
	# 	 # HiFi+ Quality.hi_res_lossless (FLAC HI_RES)
	# 	 session.audio_quality = Quality.high_lossless
	# 	 session.audio_quality = Quality.low_320k
	# 	 # album_id = "249593867"  # Alice In Chains / We Die Young (Max quality: HI_RES MHA1 SONY360)
	# 	 # album_id = "77640617"   # U2 / Achtung Baby              (Max quality: HI_RES MQA, 16bit/44100Hz)
	# 	 # album_id = "110827651"  # The Black Keys / Let's Rock    (Max quality: LOSSLESS FLAC, 24bit/48000Hz)
	# 	 album_id = "77646169"  # Beck / Sea Change               (Max quality: HI_RES_LOSSLESS FLAC, 24bit/192000Hz)
	# 	 album = session.album(album_id)
	# 	 res = album.get_audio_resolution()
	# 	 tracks = album.tracks()
	# 	 # list album tracks
	# 	 for track in tracks:
	# 		 logging.info("{}: '{}' by '{}'".format(track.id, track.name, track.artist.name))
	# 		 stream = track.get_stream()
	# 		 logging.info("MimeType:{}".format(stream.manifest_mime_type))
	#
	# 		 manifest = stream.get_stream_manifest()
	# 		 audio_resolution = stream.get_audio_resolution()
	#
	# 		 logging.info("track:{}, (quality:{}, codec:{}, {}bit/{}Hz)".format(track.id,
	# 																	 stream.audio_quality,
	# 																	 manifest.get_codecs(),
	# 																	 audio_resolution[0],
	# 																	 audio_resolution[1]))
	# 		 if stream.is_MPD:
	# 			 logging.info("MPD!")
	# 			 # HI_RES_LOSSLESS quality supported when using MPEG-DASH stream (PKCE only!)
	# 			 # 1. Export as MPD manifest
	# 			 mpd = stream.get_manifest_data()
	# 			 # 2. Export as HLS m3u8 playlist
	# 			 hls = manifest.get_hls()
	# 			 logging.info(hls)
	# 			 # with open("{}_{}.mpd".format(album_id, track.id), "w", encoding="utf-8") as my_file:
	# 			 #	my_file.write(mpd)
	# 			 # with open("{}_{}.m3u8".format(album_id, track.id), "w", encoding="utf-8") as my_file:
	# 			 #	my_file.write(hls)
	# 			 urls = manifest.get_urls()
	# 			 if urls:
	# 				 f = io.BytesIO()
	# 				 i = 0
	# 				 for url in urls:
	# 					 i += 1
	# 					 logging.info(i, end=",")
	# 					 response = requests.get(url)
	# 					 if response.status_code == 200:
	# 						 f.write(response.content)
	# 					 else:
	# 						 logging.error(f"ERROR CODE: {response.status_code}")
	# 				 f.seek(0)
	# 				 with open("test", 'wb') as a:
	# 					 a.write(f.read())
	# 				 logging.info("done")
	#
	# 		 if stream.is_BTS:
	# 			 logging.info("BTS!")
	# 			 # Direct URL (m4a or flac) is available for Quality < HI_RES_LOSSLESS
	# 			 url = manifest.get_urls()
	# 			 logging.info(f"URL = {url}")
	# 		 for url in manifest.get_urls():
	# 			 logging.info(url)
	# 		 break



"""Tauon Music Box - Configuration file module

The purpose of this module is to update, parse and write to a configuration file
"""

# Copyright © 2015-2019, Taiko2k captain(dot)gxj(at)gmail.com

#     This file is part of Tauon Music Box.
#
#     Tauon Music Box is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     Tauon Music Box is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Lesser General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with Tauon Music Box.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import logging
import os

from tauon.t_modules.t_extra import tryint


class Config:
	def __init__(self) -> None:
		self.live = []
		self.old = []

	def reset(self) -> None:
		self.live.clear()
		self.old.clear()

	def add_text(self, text: str) -> None:
		self.live.append(["comment", text])

	def add_comment(self, text: str) -> None:
		self.live.append(["comment", "# " + text])

	def br(self) -> None:
		self.live.append(["comment", ""])

	def update_value(self, key: str, value: bool | str | float) -> None:
		for item in self.live:
			if item[0] != "comment" and item[1] == key:
				item[2] = value
				break

	def load(self, path: str) -> None:
		if os.path.isfile(path):
			with open(path, encoding="utf-8") as f:
				self.old = f.readlines()

	def dump(self, path: str) -> None:
		# if os.path.exists(path) and not os.access("test.conf", os.W_OK):
		# 	logging.error("Config file cannot be written")
		# 	return

		with open(path, "w", encoding="utf-8") as f:
			for item in self.live:
				if item[0] == "comment":
					f.write(item[1])
					f.write(os.linesep)
					continue

				c = 0
				f.write(item[1])
				c += len(item[1])

				f.write(" = ")
				c += 3

				if item[0] == "bool":
					if item[2] is True:
						f.write("true")
						c += 4
					else:
						f.write("false")
						c += 5

				if item[0] == "string":
					f.write('"')
					f.write(item[2])
					f.write('"')
					c += len(item[2]) + 2

				if item[0] == "int":
					f.write(str(item[2]))
					c += len(str(item[2]))

				if item[0] == "float":
					f.write(str(item[2]))
					c += len(str(item[2]))

				if item[3]:
					d = 30 - c
					if d > 0:
						f.write(" " * d)
					f.write("  # ")
					f.write(item[3])

				f.write(os.linesep)

			f.write(os.linesep)

	def sync_add(
		self, var_type: str, key: str, default_value: bool | str | float, comment: str = ""
	) -> bool | str | float:
		got_old = False
		old_value = None

		for row in self.old:
			row = row.split(" #", 1)[0]
			if "=" in row and row.split("=", 1)[0].strip() == key:
				old_value = row.split("=", 1)[1].strip()
				if old_value:
					got_old = True
					break

		if var_type == "bool":
			if got_old:
				if old_value == "true":
					self.live.append(["bool", key, True, comment])
					return True
				if old_value == "false":
					self.live.append(["bool", key, False, comment])
					return False
			self.live.append(["bool", key, default_value, comment])
			return default_value

		if var_type == "string":
			if old_value is None:
				self.live.append(["string", key, default_value, comment])
				return default_value

			# old_value = old_value.strip('"')
			if old_value and old_value[0] == old_value[-1] == '"':
				old_value = old_value[1:-1]

			if not got_old:
				self.live.append(["string", key, default_value, comment])
				return default_value

			self.live.append(["string", key, old_value, comment])
			return old_value

			# if got_old:
			# old_value = old_value.strip('"')
			# if old_value:
			# 	self.live.append(["string", key, old_value, comment])
			# 	return old_value
			#
			# self.live.append(["string", key, default_value, comment])
			# return default_value

		if var_type == "int":
			if got_old:
				parsed_old_value = tryint(old_value)
				if isinstance(parsed_old_value, int):
					self.live.append(["int", key, parsed_old_value, comment])
					return parsed_old_value

			self.live.append(["int", key, default_value, comment])
			return default_value

		if var_type == "float":
			if got_old:
				try:
					old_value = float(old_value)
					self.live.append(["float", key, old_value, comment])
				except Exception:
					logging.exception("Warning: Config file contains invalid float")
				else:
					return old_value
			self.live.append(["float", key, default_value, comment])
			return default_value

		raise ValueError(f"var_type {var_type} is unknown and unhandled!")



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from tauon.t_modules.t_enums import MiniModeMode
from tauon.t_modules.t_extra import RadioStation

if TYPE_CHECKING:
	from pathlib import Path
	from typing import Literal


@dataclass
class Prefs:
	"""Used to hold any kind of settings"""

	# fmt:off
	view_prefs:              dict[str, bool]
	encoder_output:          Path
	window_opacity:          float
	ui_scale:                float
	power_save:              bool
	discord_allow:           bool
	left_window_control:     bool
	macstyle:                bool
	macos:                   bool
	phone:                   bool
	force_subpixel_text:     bool
	desktop:                 str
	album_mode:              bool = False
	colour_from_image:       bool = False
	dim_art:                 bool = False
	prefer_side:             bool = True  # Saves whether side panel is shown or not
	pause_fade_time:          int = 400
	change_volume_fade_time:  int = 400
	cross_fade_time:          int = 700
	volume_wheel_increment:   int = 2
	rename_folder_template:   str = "<albumartist> - <album>"
	rename_tracks_template:   str = "<tn>. <artist> - <title>.<ext>"

	enable_web:   bool = False
	allow_remote: bool = False
	expose_web:   bool = True

	enable_transcode:    bool = True
	show_rym:            bool = False
	show_band:           bool = False
	show_wiki:           bool = False
	show_transfer:       bool = True
	show_queue:          bool = True
	prefer_bottom_title: bool = True
	append_date:         bool = True

	update_title:  bool = False
	scroll_enable: bool = True
	break_enable:  bool = True

	transcode_codec:   str = "opus"
	transcode_mode:    str = "single"
	transcode_bitrate: int = 64

	#line_style: int = 1
	device:      int = 1
	device_name: str = ""

	cache_gallery:           bool = True
	gallery_row_scroll:      bool = True
	gallery_scroll_wheel_px: int = 90

	playlist_font_size:  int = 15
	playlist_row_height: int = 27

	tag_editor_name:   str = ""
	tag_editor_target: str = ""
	tag_editor_path:   str = ""
	use_lrc_instead: bool = True

	use_title:    bool = False
	auto_extract: bool = False
	auto_del_zip: bool = False
	pl_thumb:     bool = False

	use_custom_fonts:          bool = False
	linux_font:                str = "Noto Sans, Noto Sans CJK JP, Arial,"
	linux_font_semibold:       str = "Noto Sans, Noto Sans CJK JP, Arial, Medium"
	linux_font_bold:           str = "Noto Sans, Noto Sans CJK JP, Bold"
	linux_font_condensed:      str = "Noto Sans, Extra-Condensed"
	linux_font_condensed_bold: str = "Noto Sans, Extra-Condensed Bold"

	spec2_scroll: bool = True

	spec2_p_base:     list[float] = field(default_factory=lambda: [10, 10, 100])
	spec2_p_multiply: list[float] = field(default_factory=lambda: [0.5, 1, 1])

	spec2_base:           list[float] = field(default_factory=lambda: [10, 10, 100])
	spec2_multiply:       list[float] = field(default_factory=lambda: [0.5, 1, 1])
	spec2_colour_setting: str = "custom"

	auto_lfm:      bool = False
	scrobble_mark: bool = False
	enable_mpris:  bool = True

	replay_gain:       int  = 0  # 0=off 1=track 2=album
	replay_preamp:     int  = 0  # db
	radio_page_lyrics: bool = True

	show_gimage:      bool = False
	end_setting:      str  = "stop"
	show_gen:         bool = False
	show_lyrics_side: bool = True

	log_vol: bool = False

	transcode_opus_as: bool = False

	discord_active:     bool = False
	discord_ready:      bool = False
	disconnect_discord: bool = False

	monitor_downloads: bool = True
	extract_to_music:  bool = False

	enable_lb: bool = False
	lb_token:  str  = ""

	use_jump_crossfade:       bool = True
	use_transition_crossfade: bool = False
	use_pause_fade:           bool = True
	resume_on_jump:           bool = True

	show_notifications: bool = True

	true_shuffle:      bool = True
	append_total_time: bool = False
	backend:           int  = 4  # 1 basslib(removed), 2 gstreamer, 4 phazor

	album_repeat_mode:  bool = False  # passed to pctl
	album_shuffle_mode: bool = False  # passed to pctl

	finish_current: bool = False  # Finish current album when adding to queue

	reload_play_state: bool = False  # Resume playback on app restart
	resume_play_wake:  bool = False  # Resume playback on wake
	reload_state: tuple[int, float] | None = None
	# fmt:on

	mono: bool = False

	last_fm_token = None
	last_fm_username = ""

	use_card_style: bool = True

	plex_username = ""
	plex_password = ""
	plex_servername = ""

	koel_username = "admin@example.com"
	koel_password = "admin"  # noqa: S105
	koel_server_url = "http://localhost:8050"

	auto_lyrics: bool = False  # Function has been disabled
	jelly_username: str = ""
	jelly_password: str = ""
	jelly_server_url: str = "http://localhost:8096"
	jelly_timeout: int = 300  # Make configurable in case user has a large/slow setup

	auto_lyrics_checked: list = field(default_factory=list)

	show_side_art: bool = True
	always_pin_playlists: bool = True

	gallery_single_click: bool = True
	custom_bg_opacity = 40

	tabs_on_top: bool = True

	showcase_vis: bool = True
	show_lyrics_showcase: bool = True

	spec2_colour_mode = 0

	device_buffer = 80

	eq = [0.0] * 10
	use_eq: bool = False

	bio_large: bool = False
	discord_show: bool = False

	min_to_tray: bool = False

	guitar_chords: bool = False
	prefer_synced_lyrics: bool = True
	sync_lyrics_time_offset: int = -150  # in milliseconds

	playback_follow_cursor: bool = False

	gst_output = "rgvolume pre-amp=-2 fallback-gain=-6 ! autoaudiosink"

	art_bg: bool = False
	art_bg_stronger = 1
	art_bg_opacity = 10
	art_bg_blur = 9
	art_bg_always_blur: bool = False

	random_mode: bool = False
	repeat_mode: bool = False

	failed_artists: list[str] = field(default_factory=list)
	failed_background_artists: list[str] = field(default_factory=list)

	artist_list: bool = False
	auto_sort: bool = False

	transcode_inplace: bool = False

	bg_showcase_only: bool = False

	lyrics_enables: list[str] = field(default_factory=list)

	fatvap = "6b2a9499238ce6416783fc8129b8ac67"

	fanart_notify: bool = True
	discogs_pat = ""

	artist_list_prefer_album_artist: bool = True

	mini_mode_mode = MiniModeMode.MINI

	download_dir1 = ""
	dd_index: bool = False

	metadata_page_port = 7590

	custom_encoder_output = ""
	column_aa_fallback_artist: bool = False

	meta_persists_stop: bool = False
	meta_shows_selected: bool = False
	meta_shows_selected_always: bool = False

	left_align_album_artist_title: bool = False
	stop_notifications_mini_mode: bool = False
	scale_want: float = 1
	x_scale: bool = True
	hide_queue: bool = True
	show_playlist_list: bool = True
	thin_gallery_borders: bool = False

	chart_rows = 3
	chart_columns = 3
	chart_bg: list[int] = field(default_factory=lambda: [7, 7, 7])
	chart_text: bool = True
	chart_font = "Monospace 10"
	chart_tile: bool = False

	chart_cascade: bool = False
	chart_c1 = 5
	chart_c2 = 6
	chart_c3 = 10
	chart_d1 = 2
	chart_d2 = 2
	chart_d3 = 2

	art_in_top_panel: bool = True
	always_art_header: bool = False

	# center_bg = True
	ui_lang: str = "auto"
	side_panel_layout = 0
	use_absolute_track_index: bool = False

	hide_bottom_title: bool = True
	auto_goto_playing: bool = False

	diacritic_search: bool = True
	increase_gallery_row_spacing: bool = False
	center_gallery_text: bool = False

	tracklist_y_text_offset = 0
	theme: int = 7
	theme_name = "Turbo"
	transparent_mode: int = 0
	left_panel_mode = "playlist"

	folder_tree_codec_colours: bool = False

	network_stream_bitrate = 0  # 0 is off

	show_side_lyrics_art_panel: bool = True

	gst_use_custom_output: bool = False

	notify_include_album: bool = True

	auto_dl_artist_data: bool = False

	enable_fanart_artist: bool = False
	enable_fanart_bg: bool = False
	enable_fanart_cover: bool = False

	always_auto_update_playlists: bool = False

	subsonic_server = "http://localhost:4040"
	subsonic_user = ""
	subsonic_password = ""
	subsonic_password_plain: bool = False

	subsonic_playlists = {}

	write_ratings: bool = False
	rating_playtime_stars: bool = False

	lyrics_subs = {}

	radio_urls: list[RadioStation] = field(default_factory=list[RadioStation])

	lyric_metadata_panel_top: bool = False
	showcase_overlay_texture: bool = False

	sync_target = ""
	sync_deletes: bool = False
	sync_playlist: int | None = None
	download_playlist: int | None = None

	autoscan_playlist_folder: bool = False
	playlist_folder_path: str = ""

	synced_lyrics_editor_track_end_mode: Literal["stop", "autosave", "full save", "repeat"] = "repeat"
	save_lyrics_changes_to_files: bool = False

	sep_genre_multi: bool = False
	topchart_sorts_played: bool = True

	spot_client = ""
	spot_secret = ""
	spot_username = ""
	spot_password = ""
	spot_mode: bool = False
	launch_spotify_web: bool = False
	launch_spotify_local: bool = False
	remove_network_tracks: bool = False
	bypass_transcode: bool = False
	force_hide_max_button: bool = False
	zoom_art: bool = False
	auto_rec: bool = False
	radio_record_codec = "OPUS"
	pa_fast_seek: bool = False
	precache: bool = False
	# TODO(Martin): cache_list isn't really used anywhere and will always be empty?
	cache_list: list[str] = field(default_factory=list[str])
	cache_limit = 2000  # in mb
	save_window_position: bool = True
	spotify_token = ""
	always_ffmpeg: bool = False

	use_libre_fm: bool = False
	back_restarts: bool = False

	old_playlist_box_position = 0
	listenbrainz_url = ""
	maloja_enable: bool = False
	maloja_url = ""
	maloja_key = ""

	scrobble_hold: bool = False

	artist_list_sort_mode = "alpha"

	phazor_device_selected = "Default"
	phazor_devices = ["Default"]
	bg_flips = set()
	use_tray: bool = False
	tray_show_title: bool = False
	drag_to_unpin: bool = True
	enable_remote: bool = False

	artist_list_style = 1
	discord_enable: bool = False
	stop_end_queue: bool = False

	block_suspend: bool = False
	smart_bypass: bool = True
	seek_interval = 15
	shuffle_lock: bool = False
	album_shuffle_lock_mode: bool = False
	premium: bool = False
	radio_thumb_bans: list[str] = field(default_factory=list[str])
	show_nag: bool = False

	playlist_exports = {}
	show_chromecast: bool = False

	samplerate = 48000
	resample = 1
	volume_power = 2

	tmp_cache: bool = True

	sat_url = ""
	lyrics_font_size = 15

	use_gamepad: bool = True
	avoid_resampling: bool = False
	use_scancodes: bool = False

	artist_list_threshold = 4
	allow_video_formats: bool = True
	mini_mode_on_top: bool = True
	tray_theme = "pink"

	lastfm_pull_love: bool = False
	row_title_format = 1
	row_title_genre: bool = False
	row_title_separator_type = 1
	search_on_letter: bool = True

	gallery_combine_disc: bool = False
	pipewire: bool = True
	tidal_quality = 1

	jump_start: bool = True

	milk: bool = False
	auto_milk: bool = True
	loaded_preset = None

	taskbar_progress: bool = True



from enum import IntEnum


class LoaderCommand(IntEnum):
	NONE = 0
	DONE = 1
	FOLDER = 2
	FILE = 3


class PlayerState(IntEnum):
	STOPPED = 0
	PLAYING = 1
	PAUSED = 2
	URL_STREAM = 3
	SPOTIFY_MODE = 4


class PlayingState(IntEnum):
	STOPPED = 0
	PLAYING = 1
	PAUSED = 2
	URL_STREAM = 3


class StopMode(IntEnum):
	OFF = 0
	TRACK = 1
	ALBUM = 2
	TRACK_PERSIST = 3
	ALBUM_PERSIST = 4


class Backend(IntEnum):
	NONE = 0
	GSTREAMER = 2
	PHAZOR = 4


class MiniModeMode(IntEnum):
	MINI = 0
	SQUARE = 1
	LARGE = 2
	SQUARE_LARGE = 3
	TAB = 4
	SLATE = 5


class GuiMode(IntEnum):
	MAIN = 1
	MINI = 3


class QueueType(IntEnum):
	TRACK = 0
	ALBUM = 1
